import numpy as np
import pytest
from numpy.testing import assert_allclose

import quimb as qu
import quimb.tensor as qtn

dtypes = ["float32", "float64", "complex64", "complex128"]


class TestMatrixProductState:
    def test_matrix_product_state(self):
        arrays = (
            [np.random.rand(5, 2)]
            + [np.random.rand(5, 5, 2) for _ in range(3)]
            + [np.random.rand(5, 2)]
        )
        mps = qtn.MatrixProductState(arrays)
        mps.check()
        assert len(mps.tensors) == 5
        nmps = mps.reindex_sites("foo{}", inplace=False, where=slice(0, 3))
        assert nmps.site_ind_id == "k{}"
        assert isinstance(nmps, qtn.MatrixProductState)
        assert set(nmps.outer_inds()) == {"foo0", "foo1", "foo2", "k3", "k4"}
        assert set(mps.outer_inds()) == {"k0", "k1", "k2", "k3", "k4"}
        mps.site_ind_id = "foo{}"
        assert set(mps.outer_inds()) == {
            "foo0",
            "foo1",
            "foo2",
            "foo3",
            "foo4",
        }
        assert mps.site_inds == ("foo0", "foo1", "foo2", "foo3", "foo4")
        assert mps.site_ind_id == "foo{}"
        mps.show()

    @pytest.mark.parametrize(
        "dtype", [float, complex, np.complex128, np.float64, "raise"]
    )
    def test_rand_mps_dtype(self, dtype):
        if dtype == "raise":
            with pytest.raises(TypeError):
                qtn.MPS_rand_state(10, 7, dtype=dtype)
        else:
            p = qtn.MPS_rand_state(10, 7, dtype=dtype)
            assert p[0].dtype == dtype
            assert p[7].dtype == dtype

    def test_trans_invar(self):
        with pytest.raises(ValueError):
            psi = qtn.MPS_rand_state(10, 7, cyclic=False, trans_invar=True)

        psi = qtn.MPS_rand_state(10, 7, cyclic=True, trans_invar=True)
        z0 = psi.expec(psi.gate(qu.pauli("Z"), 0, contract=True))
        z3 = psi.expec(psi.gate(qu.pauli("Z"), 0, contract=True))
        z7 = psi.expec(psi.gate(qu.pauli("Z"), 0, contract=True))

        assert_allclose(z0, z3)
        assert_allclose(z3, z7)

    def test_from_dense(self):
        L = 8
        psi = qu.rand_ket(2**L)
        mps = qtn.MatrixProductState.from_dense(psi)
        assert mps.tags == qtn.oset(f"I{i}" for i in range(L))
        assert mps.site_inds == tuple(f"k{i}" for i in range(L))
        assert mps.L == L
        assert mps.bond_sizes() == [2, 4, 8, 16, 8, 4, 2]
        mpod = mps.to_qarray()
        assert qu.expec(mpod, psi) == pytest.approx(1)

    def test_from_dense_low_rank(self):
        L = 6
        psi = qu.ghz_state(L)
        mps = qtn.MatrixProductState.from_dense(psi, dims=[2] * L)
        assert mps.tags == qtn.oset(f"I{i}" for i in range(L))
        assert mps.site_inds == tuple(f"k{i}" for i in range(L))
        assert mps.L == L
        assert mps.bond_sizes() == [2, 2, 2, 2, 2]
        mpod = mps.to_qarray()
        assert qu.expec(mpod, psi) == pytest.approx(1)

    def test_left_canonize_site(self):
        a = np.random.randn(7, 2) + 1.0j * np.random.randn(7, 2)
        b = np.random.randn(7, 7, 2) + 1.0j * np.random.randn(7, 7, 2)
        c = np.random.randn(7, 2) + 1.0j * np.random.randn(7, 2)
        mps = qtn.MatrixProductState([a, b, c], site_tag_id="I{}")

        mps.left_canonize_site(0)
        assert mps["I0"].shape == (2, 2)
        assert mps["I0"].tags == qtn.oset(("I0",))
        assert mps["I1"].tags == qtn.oset(("I1",))

        U = mps["I0"].data
        assert_allclose(U.conj().T @ U, np.eye(2), atol=1e-13)
        assert_allclose(U @ U.conj().T, np.eye(2), atol=1e-13)

        # combined two site contraction is identity also
        mps.left_canonize_site(1)
        ptn = (mps.H & mps) ^ ["I0", "I1"]
        assert_allclose(ptn["I1"].data, np.eye(4), atol=1e-13)

        # try normalizing the state
        mps["I2"] /= mps["I2"].norm()

        assert_allclose(abs(mps.H @ mps), 1.0)

    def test_right_canonize_site(self):
        a = np.random.randn(7, 2) + 1.0j * np.random.randn(7, 2)
        b = np.random.randn(7, 7, 2) + 1.0j * np.random.randn(7, 7, 2)
        c = np.random.randn(7, 2) + 1.0j * np.random.randn(7, 2)
        mps = qtn.MatrixProductState([a, b, c], site_tag_id="I{}")

        mps.right_canonize_site(2)
        assert mps["I2"].shape == (2, 2)
        assert mps["I2"].tags == qtn.oset(("I2",))
        assert mps["I1"].tags == qtn.oset(("I1",))

        U = mps["I2"].data
        assert_allclose(U.conj().T @ U, np.eye(2), atol=1e-13)
        assert_allclose(U @ U.conj().T, np.eye(2), atol=1e-13)

        # combined two site contraction is identity also
        mps.right_canonize_site(1)
        ptn = (mps.H & mps) ^ ["I1", "I2"]
        assert_allclose(ptn["I1"].data, np.eye(4), atol=1e-13)

        # try normalizing the state
        mps["I0"] /= mps["I0"].norm()

        assert_allclose(mps.H @ mps, 1)

    def test_rand_mps_left_canonize(self):
        n = 10
        k = qtn.MPS_rand_state(
            n, 10, site_tag_id="foo{}", tags="bar", normalize=False
        )
        k.left_canonicalize_(normalize=True)

        assert k.count_canonized() == (9, 0)

        assert_allclose(k.H @ k, 1)
        p_tn = (k.H & k) ^ slice(0, 9)
        assert_allclose(p_tn["foo8"].data, np.eye(10), atol=1e-13)

    def test_rand_mps_left_canonize_with_bra(self):
        n = 10
        k = qtn.MPS_rand_state(
            n, 10, site_tag_id="foo{}", tags="bar", normalize=False
        )
        b = k.H
        k.left_canonicalize_(normalize=True, bra=b)
        assert_allclose(b @ k, 1)
        p_tn = (b & k) ^ slice(0, 9)
        assert_allclose(p_tn["foo8"].data, np.eye(10), atol=1e-13)

    def test_rand_mps_right_canonize(self):
        n = 10
        k = qtn.MPS_rand_state(
            n, 10, site_tag_id="foo{}", tags="bar", normalize=False
        )
        k.right_canonicalize_(normalize=True)
        assert_allclose(k.H @ k, 1)
        p_tn = (k.H & k) ^ slice(..., 0, -1)
        assert_allclose(p_tn["foo1"].data, np.eye(10), atol=1e-13)

    def test_rand_mps_right_canonize_with_bra(self):
        n = 10
        k = qtn.MPS_rand_state(
            n, 10, site_tag_id="foo{}", tags="bar", normalize=False
        )
        b = k.H
        k.right_canonicalize_(normalize=True, bra=b)
        assert_allclose(b @ k, 1)
        p_tn = (b & k) ^ slice(..., 0, -1)
        assert_allclose(p_tn["foo1"].data, np.eye(10), atol=1e-13)

    def test_rand_mps_mixed_canonize(self):
        n = 10
        rmps = qtn.MPS_rand_state(
            n, 10, site_tag_id="foo{}", tags="bar", normalize=True
        )

        # move to the center
        rmps.canonicalize_(4)
        assert rmps.count_canonized() == (4, 5)
        assert_allclose(rmps.H @ rmps, 1)
        p_tn = (rmps.H & rmps) ^ slice(0, 4) ^ slice(..., 4, -1)
        assert_allclose(p_tn["foo3"].data, np.eye(10), atol=1e-13)
        assert_allclose(p_tn["foo5"].data, np.eye(10), atol=1e-13)

        # try shifting to the right
        rmps.shift_orthogonality_center(current=4, new=8)
        assert_allclose(rmps.H @ rmps, 1)
        p_tn = (rmps.H & rmps) ^ slice(0, 8) ^ slice(..., 8, -1)
        assert_allclose(p_tn["foo7"].data, np.eye(4), atol=1e-13)
        assert_allclose(p_tn["foo9"].data, np.eye(2), atol=1e-13)

        # try shifting to the left
        rmps.shift_orthogonality_center(current=8, new=6)
        assert_allclose(rmps.H @ rmps, 1)
        p_tn = (rmps.H & rmps) ^ slice(0, 6) ^ slice(..., 6, -1)
        assert_allclose(p_tn["foo5"].data, np.eye(10), atol=1e-13)
        assert_allclose(p_tn["foo7"].data, np.eye(8), atol=1e-13)

    @pytest.mark.parametrize("dtype", dtypes)
    def test_canonize_and_calc_current_orthog_center(self, dtype):
        p = qtn.MPS_rand_state(20, 3, dtype=dtype)
        co = p.calc_current_orthog_center()
        assert co == (0, 19)
        p.canonicalize_((5, 15), co)
        co = p.calc_current_orthog_center()
        assert co == (5, 15)
        p.canonicalize_((8, 11), co)
        co = p.calc_current_orthog_center()
        assert co == (8, 11)
        assert p.dtype == dtype

    def test_can_change_data(self):
        p = qtn.MPS_rand_state(3, 10)
        assert_allclose(p.H @ p, 1)
        p[1].modify(data=np.random.randn(10, 10, 2))
        assert abs(p.H @ p - 1) > 1e-13

    def test_can_change_data_using_subnetwork(self):
        p = qtn.MPS_rand_state(3, 10)
        pH = p.H
        p.add_tag("__ket__")
        pH.add_tag("__bra__")
        tn = p | pH
        assert_allclose((tn ^ ...), 1)
        assert_allclose(
            tn[("__ket__", "I1")].data, tn[("__bra__", "I1")].data.conj()
        )
        p[1].modify(data=np.random.randn(10, 10, 2))
        assert abs((tn ^ ...) - 1) > 1e-13
        assert not np.allclose(
            tn[("__ket__", "I1")].data, tn[("__bra__", "I1")].data.conj()
        )

    def test_adding_mps(self):
        p = qtn.MPS_rand_state(10, 7)
        assert max(p["I4"].shape) == 7
        p2 = p + p
        assert max(p2["I4"].shape) == 14
        assert_allclose(p2.H @ p, 2)
        p += p
        assert max(p["I4"].shape) == 14
        assert_allclose(p.H @ p, 4)

    @pytest.mark.parametrize("method", ["svd", "eig"])
    @pytest.mark.parametrize("cutoff_mode", ["abs", "rel", "sum2"])
    def test_compress_mps(self, method, cutoff_mode):
        n = 10
        chi = 7
        p = qtn.MPS_rand_state(n, chi)
        assert max(p["I4"].shape) == chi
        p2 = p + p
        assert max(p2["I4"].shape) == chi * 2
        assert_allclose(p2.H @ p, 2)
        p2.left_compress(method=method, cutoff=1e-6, cutoff_mode=cutoff_mode)
        assert max(p2["I4"].shape) == chi
        assert_allclose(p2.H @ p, 2)
        assert p2.count_canonized() == (n - 1, 0)

    def test_compress_mps_right(self):
        p = qtn.MPS_rand_state(10, 7)
        assert max(p["I4"].shape) == 7
        p2 = p + p
        assert max(p2["I4"].shape) == 14
        assert_allclose(p2.H @ p, 2)
        p2.right_compress()
        assert max(p2["I4"].shape) == 7
        assert_allclose(p2.H @ p, 2)

    @pytest.mark.parametrize("method", ["svd", "eig"])
    def test_compress_trim_max_bond(self, method):
        p0 = qtn.MPS_rand_state(20, 20)
        p = p0.copy()
        p.compress(method=method, renorm=True)
        assert max(p["I4"].shape) == 20
        p.compress(max_bond=13, method=method, renorm=True)
        assert max(p["I4"].shape) == 13
        assert_allclose(p.H @ p, p0.H @ p0)

    def test_compress_form(self):
        p = qtn.MPS_rand_state(20, 20)
        p.compress("left")
        assert p.count_canonized() == (19, 0)
        p.compress("right")
        assert p.count_canonized() == (0, 19)
        p.compress(7)
        assert p.count_canonized() == (7, 12)
        p = qtn.MPS_rand_state(20, 20)
        p.compress("flat", absorb="left")
        assert p.count_canonized() == (0, 0)

    def test_compress_site(self):
        psi = qtn.MPS_rand_state(10, 7)
        psi.compress_site(3, max_bond=1)
        assert psi.bond_sizes() == [2, 4, 1, 1, 7, 7, 7, 4, 2]
        assert psi.calc_current_orthog_center() == (3, 3)

        psi = qtn.MPS_rand_state(10, 7)
        psi.compress_site(0, max_bond=1)
        assert psi.bond_sizes() == [1, 7, 7, 7, 7, 7, 7, 4, 2]
        assert psi.calc_current_orthog_center() == (0, 0)

        psi = qtn.MPS_rand_state(10, 7)
        psi.compress_site(9, max_bond=1)
        assert psi.bond_sizes() == [2, 4, 7, 7, 7, 7, 7, 7, 1]
        assert psi.calc_current_orthog_center() == (9, 9)

    @pytest.mark.parametrize("method", ["svd", "eig"])
    @pytest.mark.parametrize("form", ["left", "right", "raise"])
    def test_add_and_compress_mps(self, method, form):
        p = qtn.MPS_rand_state(10, 7)
        assert max(p["I4"].shape) == 7

        if form == "raise":
            with pytest.raises(ValueError):
                p.add_MPS(
                    p, compress=True, method=method, form=form, cutoff=1e-6
                )
            return

        p2 = p.add_MPS(p, compress=True, method=method, form=form, cutoff=1e-6)
        assert max(p2["I4"].shape) == 7
        assert_allclose(p2.H @ p, 2, rtol=1e-5)

    def test_subtract(self):
        a, b, c = (qtn.MPS_rand_state(10, 7) for _ in "abc")
        ab = a.H @ b
        ac = a.H @ c
        abmc = a.H @ (b - c)
        assert_allclose(ab - ac, abmc)

    def test_subtract_inplace(self):
        a, b, c = (qtn.MPS_rand_state(10, 7) for _ in "abc")
        ab = a.H @ b
        ac = a.H @ c
        b -= c
        abmc = a.H @ b
        assert_allclose(ab - ac, abmc)

    def test_amplitude(self):
        mps = qtn.MPS_rand_state(10, 7)
        k = mps.to_qarray()
        idx = np.random.randint(0, k.shape[0])
        c_b = mps.amplitude(f"{idx:0>10b}")
        assert k[idx, 0] == pytest.approx(c_b)

    def test_schmidt_values_entropy_gap_simple(self):
        n = 12
        p = qtn.MPS_rand_state(n, 16)
        p.right_canonicalize_()
        svns = []
        sgs = []
        info = {}
        for i in range(1, n):
            sgs.append(p.schmidt_gap(i, info=info))
            svns.append(p.entropy(i, info=info))

        pd = p.to_qarray()
        ex_svns = [
            qu.entropy_subsys(pd, [2] * n, range(i)) for i in range(1, n)
        ]
        ex_sgs = [qu.schmidt_gap(pd, [2] * n, range(i)) for i in range(1, n)]
        assert_allclose(ex_svns, svns)
        assert_allclose(ex_sgs, sgs)

    def test_magnetization(self):
        binary = [0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1]
        p = qtn.MPS_computational_state(binary)
        mzs = [p.magnetization(i) for i in range(len(binary))]
        assert_allclose(mzs, 0.5 - np.array(binary))

    @pytest.mark.parametrize("rescale", [False, True])
    @pytest.mark.parametrize(
        "keep", [(2, 3, 4, 6, 8), slice(-2, 4), slice(3, -1, -1), [1]]
    )
    def test_partial_trace(self, rescale, keep):
        n = 10
        p = qtn.MPS_rand_state(n, 7)
        r = p.partial_trace_to_mpo(
            keep=keep, upper_ind_id="u{}", rescale_sites=rescale
        )
        rd = r.to_qarray()
        if isinstance(keep, slice):
            keep = p.slice2sites(keep)
        else:
            if rescale:
                if keep == [1]:
                    assert r.lower_inds_present == ("u0",)
                    assert r.upper_inds_present == ("k0",)
                else:
                    assert r.lower_inds_present == (
                        "u0",
                        "u1",
                        "u2",
                        "u3",
                        "u4",
                    )
                    assert r.upper_inds_present == (
                        "k0",
                        "k1",
                        "k2",
                        "k3",
                        "k4",
                    )
            else:
                if keep == [1]:
                    assert r.lower_inds_present == ("u1",)
                    assert r.upper_inds_present == ("k1",)
                else:
                    assert r.lower_inds_present == (
                        "u2",
                        "u3",
                        "u4",
                        "u6",
                        "u8",
                    )
                    assert r.upper_inds_present == (
                        "k2",
                        "k3",
                        "k4",
                        "k6",
                        "k8",
                    )
        assert_allclose(r.trace(), 1.0)
        assert qu.isherm(rd)
        pd = p.to_qarray()
        rdd = pd.ptr([2] * n, keep=keep)
        assert_allclose(rd, rdd)

    def test_bipartite_schmidt_state(self):
        psi = qtn.MPS_rand_state(16, 5)
        psid = psi.to_qarray()
        eln = qu.logneg(psid, [2**7, 2**9])

        s_d_ket = psi.bipartite_schmidt_state(7, get="ket-dense")
        ln_d_ket = qu.logneg(s_d_ket, [5, 5])
        assert_allclose(eln, ln_d_ket, rtol=1e-5)

        s_d_rho = psi.bipartite_schmidt_state(7, get="rho-dense")
        ln_d_rho = qu.logneg(s_d_rho, [5, 5])
        assert_allclose(eln, ln_d_rho, rtol=1e-5)

        T_s_ket = psi.bipartite_schmidt_state(7, get="ket")
        assert set(T_s_ket.inds) == {"kA", "kB"}
        assert_allclose(T_s_ket.H @ T_s_ket, 1.0)

        T_s_rho = psi.bipartite_schmidt_state(7, get="rho")
        assert set(T_s_rho.outer_inds()) == {"kA", "kB", "bA", "bB"}
        assert_allclose(T_s_rho.H @ T_s_rho, 1.0)

    @pytest.mark.parametrize(
        "method", ["isvd", "svds", ("isvd", "eigsh"), ("isvd", "cholesky")]
    )
    @pytest.mark.parametrize("cyclic", [True, False])
    @pytest.mark.parametrize(
        "sysa", [range(0, 10), range(10, 20), range(20, 30), range(0, 30)]
    )
    @pytest.mark.parametrize(
        "sysb", [range(30, 40), range(40, 50), range(50, 60), range(30, 60)]
    )
    def test_partial_trace_compress(self, method, cyclic, sysa, sysb):
        k = qtn.MPS_rand_state(60, 5, cyclic=cyclic)
        kws = dict(sysa=sysa, sysb=sysb, eps=1e-6, method=method, verbosity=2)
        rhoc_ab = k.partial_trace_compress(**kws)
        assert set(rhoc_ab.outer_inds()) == {"kA", "kB", "bA", "bB"}
        inds = ["kA", "kB"], ["bA", "bB"]
        x = rhoc_ab.trace(*inds)
        assert_allclose(1.0, x, rtol=1e-3)

    @pytest.mark.parametrize("cyclic", [True, False])
    def test_known_bad_case(self, cyclic):
        k = qtn.MPS_rand_state(5, 10, cyclic=cyclic)
        rhoc_ab = k.partial_trace_compress(sysa=range(2), sysb=range(2, 4))
        inds = ["kA", "kB"], ["bA", "bB"]
        x = rhoc_ab.trace(*inds)
        assert_allclose(1.0, x, rtol=1e-3)

    @pytest.mark.parametrize(
        "block",
        [
            0,
            20,
            39,
            slice(0, 5),
            slice(20, 25),
            slice(35, 40),
            slice(38, 42),
            slice(-3, 2),
        ],
    )
    @pytest.mark.parametrize("dtype", [float, complex])
    def test_canonize_cyclic(self, dtype, block):
        k = qtn.MPS_rand_state(40, 10, dtype=dtype, cyclic=True)
        b = k.H
        k.add_tag("KET")
        b.add_tag("BRA")
        kb = b | k

        assert not np.allclose(k[block].H @ k[block], 1.0)
        assert not np.allclose(b[block].H @ b[block], 1.0)
        k.canonize_cyclic(block, bra=b)
        assert_allclose(k[block].H @ k[block], 1.0, rtol=2e-4)
        assert_allclose(b[block].H @ b[block], 1.0, rtol=2e-4)

        ii = kb.select(block, which="!any") ^ all

        if isinstance(block, slice):
            start, stop = block.start, block.stop
        else:
            start, stop = block, block + 1

        assert len(kb.select_tensors(block, "any")) == 2 * (stop - start)

        (ul,) = qtn.bonds(
            kb[k.site_tag(start - 1), "BRA"], kb[k.site_tag(start), "BRA"]
        )
        (ur,) = qtn.bonds(
            kb[k.site_tag(stop - 1), "BRA"], kb[k.site_tag(stop), "BRA"]
        )
        (ll,) = qtn.bonds(
            kb[k.site_tag(start - 1), "KET"], kb[k.site_tag(start), "KET"]
        )
        (lr,) = qtn.bonds(
            kb[k.site_tag(stop - 1), "KET"], kb[k.site_tag(stop), "KET"]
        )

        ii = ii.to_qarray((ul, ur), (ll, lr))
        assert_allclose(ii, np.eye(ii.shape[0]), rtol=0.001, atol=0.001)

    @pytest.mark.parametrize("bsz", [1, 2])
    @pytest.mark.parametrize("propagate_tags", [False, True])
    @pytest.mark.parametrize("contract", [False, True])
    def test_gate_no_contract(self, bsz, propagate_tags, contract):
        p = qtn.MPS_rand_state(5, 7, tags={"PSI0"})
        q = p.copy()
        G = qu.rand_uni(2**bsz)
        p = p.gate_(
            G,
            where=[i for i in range(2, 2 + bsz)],
            tags="G",
            contract=contract,
            propagate_tags=propagate_tags,
        )
        TG = p["G"]
        if propagate_tags or contract:
            assert p.site_tag(2) in TG.tags
        assert ("PSI0" in TG.tags) == (propagate_tags is True) or contract
        assert (p.H & p) ^ all == pytest.approx(1.0)
        assert abs((q.H & p) ^ all) < 1.0
        assert len(p.tensors) == 6 - int(contract) * bsz
        assert set(p.outer_inds()) == {f"k{i}" for i in range(5)}

    @pytest.mark.parametrize(
        "propagate_tags", [False, "sites", "register", True]
    )
    def test_gate_split_gate(self, propagate_tags):
        p = qtn.MPS_rand_state(5, 7, tags={"PSI0"})
        q = p.copy()
        G = qu.CNOT()
        p = p.gate_(
            G,
            where=[i for i in range(2, 4)],
            tags="G",
            contract="split-gate",
            propagate_tags=propagate_tags,
        )

        TG = sorted(p["G"], key=lambda t: sorted(t.tags))

        if propagate_tags is False:
            assert TG[0].tags == qtn.oset(("G",))
            assert TG[1].tags == qtn.oset(("G",))

        elif propagate_tags == "register":
            assert TG[0].tags == qtn.oset(["G", "I2"])
            assert TG[1].tags == qtn.oset(["G", "I3"])

        elif propagate_tags == "sites":
            assert TG[0].tags == qtn.oset(["G", "I2", "I3"])
            assert TG[1].tags == qtn.oset(["G", "I2", "I3"])

        elif propagate_tags is True:
            assert TG[0].tags == qtn.oset(["PSI0", "G", "I2", "I3"])
            assert TG[1].tags == qtn.oset(["PSI0", "G", "I2", "I3"])

        assert (p.H & p) ^ all == pytest.approx(1.0)
        assert abs((q.H & p) ^ all) < 1.0
        assert len(p.tensors) == 7
        assert set(p.outer_inds()) == {f"k{i}" for i in range(5)}

    def test_gate_swap_and_split_bond_sizes(self):
        n = 10
        p = qtn.MPS_computational_state("0" * n)
        assert p.bond_sizes() == [1] * (n - 1)
        G = qu.rand_uni(4)
        p.gate_(G, (1, n - 2), contract="swap+split")
        assert p.bond_sizes() == [1] + [2] * (n - 3) + [1]

    def test_gate_swap_and_split_matches(self):
        k = qtn.MPS_rand_state(6, 7)
        kr = k.copy()

        gates = [qu.rand_uni(4) for _ in range(3)]
        wheres = [(0, 5), (3, 2), (4, 1)]

        for G, (i, j) in zip(gates, wheres):
            k.gate_(G, (i, j), contract="swap+split")
            kr.gate_inds_(G, (k.site_ind(i), k.site_ind(j)))

        assert_allclose(k.to_dense(), kr.to_dense())

    def test_flip(self):
        p = qtn.MPS_rand_state(5, 3)
        pf = p.flip()
        # we want a single index per dimension, not all combined into one
        inds = [[ix] for ix in p.site_inds]
        assert_allclose(p.to_qarray(*inds), pf.to_qarray(*inds).transpose())

    def test_correlation(self):
        ghz = (
            qtn.MPS_computational_state("0000")
            + qtn.MPS_computational_state("1111")
        ) / 2**0.5

        assert ghz.correlation(qu.pauli("Z"), 0, 1) == pytest.approx(1.0)
        assert ghz.correlation(qu.pauli("Z"), 1, 2) == pytest.approx(1.0)
        assert ghz.correlation(qu.pauli("Z"), 3, 1) == pytest.approx(1.0)
        assert ghz.correlation(
            qu.pauli("Z"), 3, 1, B=qu.pauli("Y")
        ) == pytest.approx(0.0)

        assert ghz.H @ ghz == pytest.approx(1.0)

    def test_gate_split(self):
        psi = qtn.MPS_rand_state(10, 3)
        psi2 = psi.copy()
        G = qu.eye(2) & qu.eye(2)
        psi.gate_split_(G, (2, 3), cutoff=0)
        assert psi.bond_size(2, 3) == 6
        assert psi.H @ psi2 == pytest.approx(1.0)

        # check a unitary application
        G = qu.rand_uni(2**2)
        psi.gate_split_(G, (7, 8))
        psi.compress()
        assert psi.bond_size(2, 3) == 3
        assert psi.bond_size(7, 8) > 3
        assert psi.H @ psi == pytest.approx(1.0)
        assert abs(psi2.H @ psi) < 1.0

        # check matches dense application of gate
        psid = psi2.to_qarray()
        Gd = qu.ikron(G, [2] * 10, (7, 8))
        assert psi.to_qarray().H @ (Gd @ psid) == pytest.approx(1.0)

    def test_swap_far_sites(self):
        psi = qtn.MPS_rand_state(7, 2)
        for i, j in [(0, 6), (6, 1), (5, 2)]:
            k1 = psi.to_qarray(
                [
                    psi.site_ind(j if site == i else i if site == j else site)
                    for site in psi.sites
                ]
            )
            k2 = psi.swap_sites_with_compress(i, j).to_qarray()
            assert qu.fidelity(k1, k2) == pytest.approx(1.0)

    def test_swap_gating(self):
        psi0 = qtn.MPS_rand_state(20, 5)
        CNOT = qu.controlled("not")
        psi0XX = psi0.gate(CNOT, (4, 13))
        psi0XX_s = psi0.gate_with_auto_swap(CNOT, (4, 13))
        assert psi0XX.H @ psi0XX_s == pytest.approx(1.0)

    def test_auto_split_detection(self):
        psi0 = qtn.MPS_computational_state("00")
        CNOT = qu.controlled("not")
        ISWAP = qu.iswap()
        G = qu.rand_uni(4)

        opts = {"contract": "auto-split-gate", "where": (0, 1)}

        psi_cnot = psi0.gate(CNOT, **opts)
        psi_iswap = psi0.gate(ISWAP, **opts)
        psi_G = psi0.gate(G, **opts)

        assert (
            psi_cnot.max_bond()
            == psi_iswap.max_bond()
            == psi_G.max_bond()
            == 2
        )

        assert len(psi_cnot.tensors) == len(psi_iswap.tensors) == 4
        assert len(psi_G.tensors) == 3

    @pytest.mark.parametrize("cur_orthog", (None, 3))
    @pytest.mark.parametrize("site", (0, 5, 9))
    @pytest.mark.parametrize("outcome", (None, 2))
    @pytest.mark.parametrize("renorm", (True, False))
    @pytest.mark.parametrize("remove", (True, False))
    def test_mps_measure(self, cur_orthog, site, outcome, renorm, remove):
        psi = qtn.MPS_rand_state(10, 7, phys_dim=3, dtype=complex)
        if cur_orthog:
            psi.canonicalize_(cur_orthog)
        outcome, psim = psi.measure(
            site,
            outcome=outcome,
            cur_orthog=cur_orthog,
            renorm=renorm,
            remove=remove,
        )
        newL = 10 - int(remove)
        assert psim.L == newL
        assert psim.num_tensors == newL
        assert set(psim.site_tags) == {f"I{i}" for i in range(newL)}
        assert set(psim.site_inds) == {f"k{i}" for i in range(newL)}
        if renorm:
            assert psim.H @ psim == pytest.approx(1.0)
        else:
            assert 0.0 < psim.H @ psim < 1.0
        new_can_cen = min(site, newL - 1)
        t = psim[new_can_cen]
        if renorm:
            assert t.H @ t == pytest.approx(1.0)
        else:
            0.0 < t.H @ t < 1.0

    def test_measure_known_outcome(self):
        mps = qtn.MPS_computational_state("010101")
        assert mps.measure_(3, get="outcome") == 1

    def test_permute_arrays(self):
        mps = qtn.MPS_rand_state(7, 5)
        k0 = mps.to_qarray()
        mps.canonicalize_(3)
        mps.permute_arrays("prl")
        assert mps[0].shape == (2, 2)
        assert mps[1].shape == (2, 4, 2)
        assert mps[2].shape == (2, 5, 4)
        kf = mps.to_qarray()
        assert qu.fidelity(k0, kf) == pytest.approx(1.0)

    @pytest.mark.parametrize("where", [(7, 2, 4), (0, 1), (6, 7), (0, 3, 7)])
    @pytest.mark.parametrize("phys_dim", [2, 3])
    def test_gate_non_local(self, where, phys_dim):
        psi = qtn.MPS_rand_state(8, 3, phys_dim=phys_dim, dtype="complex128")
        G = qu.rand_uni(phys_dim ** len(where))
        Gpsi = psi.gate_nonlocal(G, where=where)
        assert Gpsi.H @ Gpsi == pytest.approx(1.0)
        assert Gpsi.distance_normalized(
            psi.gate(G, where, contract=False)
        ) == pytest.approx(0.0, abs=1e-6)

    def test_sample_configuration(self):
        psi = qtn.MPS_rand_state(10, 7)
        config, omega = psi.sample_configuration()
        assert len(config) == 10
        assert abs(
            psi.isel(
                {psi.site_ind(i): xi for i, xi in enumerate(config)}
            ).contract()
        ) ** 2 == pytest.approx(omega)

    def test_sample_seed(self):
        psi = qtn.MPS_rand_state(10, 7)
        configs = [
            "".join(map(str, config))
            for config, _ in psi.sample(10, seed=1234)
        ]
        assert len(set(configs)) > 1

    def test_compute_local_expectation(self):
        psi = qtn.MPS_rand_state(10, 7, dtype="complex128")
        terms = {(i, i + 1): qu.rand_herm(4) for i in range(9)}

        ex = psi.compute_local_expectation_exact(terms)
        xa = psi.compute_local_expectation(terms, method="canonical")
        assert xa == pytest.approx(ex)
        xb = psi.compute_local_expectation(terms, method="envs")
        assert xb == pytest.approx(ex)

    def test_single_site_constructor(self):
        arrays = [np.random.randn(2)]
        mps = qtn.MatrixProductState(arrays)
        mps.check()
        assert mps.L == 1
        assert mps.num_tensors == 1
        assert mps.num_indices == 1
        assert not mps.cyclic

        arrays = [np.random.randn(3, 3, 2)]
        mps = qtn.MatrixProductState(arrays)
        mps.check()
        assert mps.L == 1
        assert mps.num_tensors == 1
        assert mps.num_indices == 2
        assert mps.cyclic


class TestMatrixProductOperator:
    @pytest.mark.parametrize("cyclic", [False, True])
    def test_matrix_product_operator(self, cyclic):
        end_shape = (5, 5, 2, 2) if cyclic else (5, 2, 2)

        tensors = (
            [np.random.rand(*end_shape)]
            + [np.random.rand(5, 5, 2, 2) for _ in range(3)]
            + [np.random.rand(*end_shape)]
        )
        mpo = qtn.MatrixProductOperator(tensors)

        mpo.show()
        assert len(mpo.tensors) == 5
        assert mpo.upper_inds == ("k0", "k1", "k2", "k3", "k4")
        assert mpo.lower_inds == ("b0", "b1", "b2", "b3", "b4")
        op = mpo ^ ...
        # this would rely on left to right contraction if not in set form
        assert set(op.inds) == {
            "k0",
            "b0",
            "k1",
            "b1",
            "k2",
            "b2",
            "k3",
            "b3",
            "k4",
            "b4",
        }

        assert set(mpo.site_tags) == {f"I{i}" for i in range(5)}
        assert all(f"I{i}" in mpo.tags for i in range(5))
        mpo.site_tag_id = "TEST1,{}"
        assert set(mpo.site_tags) == {f"TEST1,{i}" for i in range(5)}
        assert not any(f"I{i}" in mpo.tags for i in range(5))
        assert all(f"TEST1,{i}" in mpo.tags for i in range(5))

    @pytest.mark.parametrize("cyclic", [False, True])
    def test_compress_mpo(self, cyclic):
        A = qtn.MPO_rand(12, 5, cyclic=cyclic)
        assert all(b == 5 for b in A.bond_sizes())
        A.expand_bond_dimension(10)
        assert all(b == 10 for b in A.bond_sizes())
        A.compress()
        assert all(b in (4, 5) for b in A.bond_sizes())

    def test_add_mpo(self):
        h = qtn.MPO_rand_herm(12, 5)
        h2 = h + h
        assert max(h2[6].shape) == 10
        t = h.trace()
        t2 = h2.trace()
        assert_allclose(2 * t, t2)

    @pytest.mark.parametrize("p", [2, 3, 4])
    def test_add_mpo_length_1(self, p):
        mpoa = qtn.MPO_rand(L=1, bond_dim=1, phys_dim=p)
        mpob = qtn.MPO_rand(L=1, bond_dim=1, phys_dim=p)
        mpoc = mpoa + mpob
        assert mpoc[0].shape == (p, p)
        assert_allclose(mpoc.to_dense(), mpoa.to_dense() + mpob.to_dense())

    def test_adding_mpo(self):
        h = qtn.MPO_ham_heis(6)
        hd = h.to_qarray()
        assert_allclose(h @ h.H, (hd @ hd.H).tr())
        h2 = h + h
        assert_allclose(h2 @ h2.H, (hd @ hd.H).tr() * 4)
        h2.right_compress()
        assert_allclose(h2 @ h2.H, (hd @ hd.H).tr() * 4)
        assert max(h2["I3"].shape) == 5

    @pytest.mark.parametrize("cyclic", (False, True))
    def test_subtract_mpo(self, cyclic):
        a, b = (
            qtn.MPO_rand(13, 7, cyclic=cyclic),
            qtn.MPO_rand(13, 7, cyclic=cyclic),
        )
        x1 = a.trace() - b.trace()
        assert_allclose(x1, (a - b).trace())
        a -= b
        assert_allclose(x1, a.trace())

    @pytest.mark.parametrize("cyclic", (False, True))
    @pytest.mark.parametrize("rand_strength", (0, 1e-9))
    def test_expand_mpo(self, cyclic, rand_strength):
        h = qtn.MPO_ham_heis(12, cyclic=cyclic)
        assert h[0].dtype == float
        he = h.expand_bond_dimension(13, rand_strength=rand_strength)
        assert h[0].dtype == float
        assert max(he[6].shape) == 13

        if cyclic:
            assert he.bond_size(0, -1) == 13

        t = h.trace()
        te = he.trace()
        assert_allclose(t, te)

    @pytest.mark.parametrize("cyclic", (False, True))
    @pytest.mark.parametrize("rand_strength", (0, 1e-9))
    def test_expand_mpo_limited(self, cyclic, rand_strength):
        h = qtn.MPO_ham_heis(12, cyclic=cyclic)
        he = h.expand_bond_dimension(3, rand_strength=rand_strength)
        # should do nothing
        assert max(he[6].shape) == 5

    def test_mpo_identity(self):
        k = qtn.MPS_rand_state(13, 7)
        b = qtn.MPS_rand_state(13, 7)
        o1 = k @ b
        i = qtn.MPO_identity(13)
        k, i, b = qtn.tensor_network_align(k, i, b)
        o2 = (k & i & b) ^ ...
        assert_allclose(o1, o2)

    @pytest.mark.parametrize("cyclic", [False, True])
    @pytest.mark.parametrize("dtype", (complex, float))
    def test_mpo_rand_herm_and_trace(self, dtype, cyclic):
        op = qtn.MPO_rand_herm(
            20, bond_dim=5, phys_dim=3, dtype=dtype, cyclic=cyclic
        )
        assert_allclose(op.H @ op, 1.0)
        tr_val = op.trace()
        assert tr_val != 0.0
        assert_allclose(tr_val.imag, 0.0, atol=1e-14)

    @pytest.mark.parametrize("cyclic", [False, True])
    def test_mpo_rand_herm_trace_and_identity_like(self, cyclic):
        op = qtn.MPO_rand_herm(
            20, bond_dim=5, phys_dim=3, upper_ind_id="foo{}", cyclic=cyclic
        )
        t = op.trace()
        assert t != 0.0
        Id = qtn.MPO_identity_like(op)
        assert_allclose(Id.trace(), 3**20)
        Id[0] *= 3 / 3**20
        op += Id
        assert_allclose(op.trace(), t + 3)

    def test_partial_transpose(self):
        p = qtn.MPS_rand_state(8, 10)
        r = p.partial_trace_to_mpo([2, 3, 4, 5, 6, 7])
        rd = r.to_qarray()

        assert qu.isherm(rd)
        assert qu.ispos(rd)

        rpt = r.partial_transpose([0, 1, 2])
        rptd = rpt.to_qarray()

        upper_inds = tuple(f"b{i}" for i in range(6))
        lower_inds = tuple(f"k{i}" for i in range(6))
        outer_inds = rpt.outer_inds()
        assert all(i in outer_inds for i in upper_inds + lower_inds)

        assert qu.isherm(rptd)
        assert not qu.ispos(rptd)

    def test_upper_lower_ind_id_guard(self):
        A = qtn.MPO_rand(8, 5)
        with pytest.raises(ValueError):
            A.upper_ind_id = "b{}"
        with pytest.raises(ValueError):
            A.lower_ind_id = "k{}"

    @pytest.mark.parametrize("cyclic", (False, True))
    def test_apply_mpo(self, cyclic):
        A = qtn.MPO_rand(8, 5, cyclic=cyclic)
        B = qtn.MPO_rand(
            8, 5, upper_ind_id="q{}", lower_ind_id="w{}", cyclic=cyclic
        )
        C = A.apply(B)
        assert C.max_bond() == 25
        assert C.upper_ind_id == "q{}"
        assert C.lower_ind_id == "w{}"
        Ad, Bd, Cd = A.to_qarray(), B.to_qarray(), C.to_qarray()
        assert_allclose(Ad @ Bd, Cd)

    @pytest.mark.parametrize("cyclic", (False, True))
    @pytest.mark.parametrize("site_ind_id", ("k{}", "test{}"))
    def test_apply_mps(self, cyclic, site_ind_id):
        A = qtn.MPO_rand(8, 5, cyclic=cyclic)
        x = qtn.MPS_rand_state(8, 4, site_ind_id=site_ind_id, cyclic=cyclic)
        y = A.apply(x)
        assert y.max_bond() == 20
        assert isinstance(y, qtn.MatrixProductState)
        assert len(y.tensors) == 8
        assert y.site_ind_id == site_ind_id
        Ad, xd, yd = A.to_qarray(), x.to_qarray(), y.to_qarray()
        assert_allclose(Ad @ xd, yd)

    def test_permute_arrays(self):
        mpo = qtn.MPO_rand(4, 3)
        A0 = mpo.to_qarray()
        mpo.permute_arrays("drul")
        assert mpo[0].shape == (2, 3, 2)
        assert mpo[1].shape == (2, 3, 2, 3)
        Af = mpo.to_qarray()
        assert_allclose(A0, Af)

    def test_from_dense(self):
        A = qu.rand_uni(2**4)
        mpo = qtn.MatrixProductOperator.from_dense(A)
        assert mpo.L == 4
        assert_allclose(A, mpo.to_dense())

    def test_from_dense_sites(self):
        dims = [2, 3, 4, 5]
        A = qu.rand_uni(2 * 3 * 4 * 5)
        sites = [3, 1, 0, 2]
        mpo = qtn.MatrixProductOperator.from_dense(A, dims, sites=sites)
        assert mpo.L == 4
        perm = [sites.index(i) for i in range(4)]
        assert_allclose(qu.permute(A, dims, perm), mpo.to_dense())

    def test_fill_empty_sites(self):
        mps = qtn.MPS_rand_state(7, 3)
        k = mps.to_dense()
        A, B, C = (qu.rand_uni(2) for _ in range(3))
        Ak = qu.ikron((A, B, C), [2] * 7, [5, 2, 3]) @ k

        ABC = A & B & C
        mpo = qtn.MatrixProductOperator.from_dense(ABC, sites=[5, 2, 3], L=7)
        assert mpo.bond_size(2, 3) == 1
        assert mpo.num_tensors == 3
        assert mpo[3].bonds(mpo[5])
        mpo.fill_empty_sites_("minimal")
        assert not mpo[3].bonds(mpo[5])
        assert mpo.num_tensors == 4
        assert_allclose(
            mps.gate_with_op_lazy(mpo).to_dense(),
            Ak,
        )
        mpo.fill_empty_sites_("full")
        assert mpo.num_tensors == 7
        assert_allclose(
            mps.gate_with_op_lazy(mpo).to_dense(),
            Ak,
        )

    def test_single_site_constructor(self):
        arrays = [np.random.randn(2, 2)]
        mpo = qtn.MatrixProductOperator(arrays)
        mpo.check()
        assert mpo.L == 1
        assert mpo.num_tensors == 1
        assert mpo.num_indices == 2
        assert not mpo.cyclic

        arrays = [np.random.randn(3, 3, 2, 2)]
        mpo = qtn.MatrixProductOperator(arrays)
        mpo.check()
        assert mpo.L == 1
        assert mpo.num_tensors == 1
        assert mpo.num_indices == 3
        assert mpo.cyclic


# --------------------------------------------------------------------------- #
#                         Test specific 1D instances                          #
# --------------------------------------------------------------------------- #


class TestSpecificStatesOperators:
    @pytest.mark.parametrize("cyclic", [False, True])
    def test_rand_ket_mps(self, cyclic):
        n = 10
        rmps = qtn.MPS_rand_state(
            n, 10, site_tag_id="foo{}", tags="bar", cyclic=cyclic
        )
        assert rmps[0].tags == qtn.oset(["foo0", "bar"])
        assert rmps[3].tags == qtn.oset(["foo3", "bar"])
        assert rmps[-1].tags == qtn.oset(["foo9", "bar"])

        rmpsH_rmps = rmps.H & rmps
        assert len(rmpsH_rmps.tag_map["foo0"]) == 2
        assert len(rmpsH_rmps.tag_map["bar"]) == n * 2

        assert_allclose(rmps.H @ rmps, 1)
        c = (rmps.H & rmps) ^ slice(0, 5) ^ slice(9, 4, -1) ^ slice(4, 6)
        assert_allclose(c, 1)

        assert rmps[0].data.ndim == (3 if cyclic else 2)
        assert rmps[-1].data.ndim == (3 if cyclic else 2)

    def test_mps_computation_state(self):
        p = qtn.MPS_neel_state(10)
        pd = qu.neel_state(10)
        assert_allclose(p.to_qarray(), pd)

    def test_zero_state(self):
        z = qtn.MPS_zero_state(21, 7)
        p = qtn.MPS_rand_state(21, 13)
        assert_allclose(p.H @ z, 0.0)
        assert_allclose(p.H @ p, 1.0)
        zp = z + p
        assert max(zp[13].shape) == 20
        assert_allclose(zp.H @ p, 1.0)

    @pytest.mark.parametrize("cyclic", [False, True])
    @pytest.mark.parametrize("j", [7 / 11, 1, (0.2, 0.3, 0.4)])
    @pytest.mark.parametrize("bz", [0, 7 / 11, 1])
    @pytest.mark.parametrize("n", [2, 3, 4])
    def test_mpo_site_ham_heis(self, cyclic, j, bz, n):
        hh_mpo = qtn.MPO_ham_heis(n, tags=["foo"], cyclic=cyclic, j=j, bz=bz)
        assert hh_mpo[0].tags == qtn.oset(["I0", "foo"])
        assert hh_mpo[1].tags == qtn.oset(["I1", "foo"])
        assert hh_mpo[-1].tags == qtn.oset([f"I{n - 1}", "foo"])
        assert hh_mpo.shape == (2,) * 2 * n
        hh_ex = qu.ham_heis(n, cyclic=cyclic, j=j, b=bz)
        assert_allclose(
            qu.eigvalsh(hh_ex), qu.eigvalsh(hh_mpo.to_qarray()), atol=1e-13
        )

    def test_mpo_zeros(self):
        mpo0 = qtn.MPO_zeros(10)
        assert mpo0.trace() == 0.0
        assert mpo0.H @ mpo0 == 0.0

    @pytest.mark.parametrize("cyclic", (False, True))
    def test_mpo_zeros_like(self, cyclic):
        A = qtn.MPO_rand(10, 7, phys_dim=3, normalize=False, cyclic=cyclic)
        Z = qtn.MPO_zeros_like(A)
        assert A @ Z == 0.0
        assert Z.cyclic == cyclic
        x1 = A.trace()
        x2 = (A + Z).trace()
        assert_allclose(x1, x2)


class TestDense1D:
    def test_simple(self):
        n = 10
        d_psi = qu.computational_state("0" * n)

        t_psi = qtn.Dense1D(d_psi)
        assert set(t_psi.outer_inds()) == {f"k{i}" for i in range(n)}
        assert t_psi.tags == qtn.oset(f"I{i}" for i in range(n))

        for i in range(n):
            assert t_psi.H @ t_psi.gate(qu.pauli("Z"), i) == pytest.approx(1)

        for i in range(n):
            t_psi.gate_(qu.hadamard(), i)

        assert len(t_psi.tensors) == n + 1

        # should have '++++++++++'
        assert t_psi.H @ t_psi == pytest.approx(1)
        for i in range(n):
            assert t_psi.H @ t_psi.gate(qu.pauli("X"), i) == pytest.approx(1)

    def test_rand(self):
        t_psi = qtn.Dense1D.rand(7, dtype="complex64")
        assert t_psi.shape == (2,) * 7
        assert t_psi.dtype == "complex64"
        assert (t_psi.H @ t_psi) == pytest.approx(1.0)


class TestTensor1DCompress:
    @pytest.mark.parametrize(
        "method",
        [
            "direct",
            "dm",
            "fit",
            "zipup",
            "zipup-first",
        ],
    )
    @pytest.mark.parametrize("dtype", dtypes)
    def test_mps_partial_mpo_apply(self, method, dtype):
        mps = qtn.MPS_rand_state(10, 7, dtype=dtype)
        A = qu.rand_uni(2**3, dtype=dtype)
        where = [8, 4, 5]
        mpo = qtn.MatrixProductOperator.from_dense(A, sites=where)
        new = mps.gate_with_op_lazy(mpo)
        assert (
            qtn.tensor_network_1d_compress(new, method=method, inplace=True)
            is new
        )
        assert new.num_tensors == 10
        assert new.distance_normalized(mps.gate(A, where)) == pytest.approx(
            0.0, abs=1e-3 if dtype in ("float32", "complex64") else 1e-6
        )

    @pytest.mark.parametrize(
        "method",
        [
            "direct",
            "dm",
            "fit",
            "zipup",
            "zipup-first",
            "src",
            "src-first",
        ],
    )
    @pytest.mark.parametrize("sweep_reverse", [False, True])
    def test_mpo_compress_opts(self, method, sweep_reverse):
        L = 6
        A = qtn.MPO_rand(L, 2, phys_dim=3, tags="A")
        B = qtn.MPO_rand(L, 3, phys_dim=3, tags="B")
        AB = A.gate_upper_with_op_lazy(B)
        assert AB.num_tensors == 2 * L
        ABc = qtn.tensor_network_1d_compress(
            AB,
            method=method,
            max_bond=5,
            cutoff=1e-6,
            sweep_reverse=sweep_reverse,
            inplace=False,
        )
        assert ABc.num_tensors == L
        assert ABc.num_indices == 2 * L + L - 1
        assert ABc.max_bond() == 5
        if sweep_reverse:
            assert ABc.calc_current_orthog_center() == (L - 1, L - 1)
        else:
            assert ABc.calc_current_orthog_center() == (0, 0)

        for site in range(L):
            assert set(ABc[site].tags) == {"A", "B", f"I{site}"}



import numpy as np
import pytest
from numpy.testing import assert_allclose

import quimb as qu
import quimb.tensor as qtn


class TestGeometries:
    @pytest.mark.parametrize("cyclic", [False, True])
    @pytest.mark.parametrize(
        "edge_fn,shape,percell,coordination",
        [
            (qtn.edges_2d_square, (3, 3), 1, 4),
            (qtn.edges_2d_hexagonal, (3, 3), 2, 3),
            (qtn.edges_2d_kagome, (3, 3), 3, 4),
            (qtn.edges_2d_triangular, (3, 3), 1, 6),
            (qtn.edges_2d_triangular_rectangular, (3, 3), 2, 6),
            (qtn.edges_3d_cubic, (3, 3, 3), 1, 6),
            (qtn.edges_3d_pyrochlore, (3, 3, 3), 4, 6),
            (qtn.edges_3d_diamond, (3, 3, 3), 2, 4),
            (qtn.edges_3d_diamond_cubic, (2, 2, 2), 8, 4),
        ],
    )
    def test_basic(self, cyclic, edge_fn, shape, percell, coordination):
        edges = edge_fn(*shape, cyclic=cyclic)
        tn = qtn.TN_rand_from_edges(edges, D=2)
        assert tn.num_tensors == qu.prod(shape) * percell
        assert max(t.ndim for t in tn) == coordination


class TestSpinHam1D:
    @pytest.mark.parametrize("cyclic", [False, True])
    def test_var_terms(self, cyclic):
        n = 8
        Hd = qu.ham_mbl(n, dh=0.77, seed=42, cyclic=cyclic)
        Ht = qtn.MPO_ham_mbl(n, dh=0.77, seed=42, cyclic=cyclic).to_dense()
        assert_allclose(Hd, Ht)

    @pytest.mark.parametrize("var_two", ["none", "some", "only"])
    @pytest.mark.parametrize(
        "var_one", ["some", "only", "only-some", "def-only", "none"]
    )
    def test_specials(self, var_one, var_two):
        K1 = qu.rand_herm(2**1)

        n = 10
        HB = qtn.SpinHam1D(S=1 / 2)

        if var_two == "some":
            HB += 1, K1, K1
            HB[4, 5] += 1, K1, K1
            HB[7, 8] += 1, K1, K1
        elif var_two == "only":
            for i in range(n - 1):
                HB[i, i + 1] += 1, K1, K1
        else:
            HB += 1, K1, K1

        if var_one == "some":
            HB += 1, K1
            HB[2] += 1, K1
            HB[3] += 1, K1
        elif var_one == "only":
            for i in range(n - 1):
                HB[i] += 1, K1
        elif var_one == "only-some":
            HB[1] += 1, K1
        elif var_one == "def-only":
            HB += 1, K1

        HB.build_local_ham(n)
        H_mpo = HB.build_mpo(n)
        H_sps = HB.build_sparse(n)

        assert_allclose(H_mpo.to_dense(), H_sps.toarray())

    def test_no_default_term(self):
        N = 10
        builder = qtn.SpinHam1D(1 / 2)

        for i in range(N - 1):
            builder[i, i + 1] += 1.0, "Z", "Z"

        H = builder.build_mpo(N)

        dmrg = qtn.DMRG2(H)
        dmrg.solve(verbosity=1)

        assert dmrg.energy == pytest.approx(-2.25)


class TestMPSSpecificStates:
    def test_site_varying_phys_dim(self):
        k = qtn.MPS_rand_state(5, 4, phys_dim=[2, 3, 3, 2, 5])
        assert k.H @ k == pytest.approx(1.0)
        assert k.outer_dims_inds() == (
            (2, "k0"),
            (3, "k1"),
            (3, "k2"),
            (2, "k3"),
            (5, "k4"),
        )

    @pytest.mark.parametrize("dtype", ["float32", "complex64"])
    def test_ghz_state(self, dtype):
        mps = qtn.MPS_ghz_state(5, dtype=dtype)
        assert mps.dtype == dtype
        psi = qu.ghz_state(5, dtype=dtype)
        assert mps.H @ mps == pytest.approx(1.0)
        assert mps.bond_sizes() == [2, 2, 2, 2]
        assert qu.fidelity(psi, mps.to_dense()) == pytest.approx(1.0)

    @pytest.mark.parametrize("dtype", ["float32", "complex64"])
    def test_w_state(self, dtype):
        mps = qtn.MPS_w_state(5, dtype=dtype)
        assert mps.dtype == dtype
        psi = qu.w_state(5, dtype=dtype)
        assert mps.H @ mps == pytest.approx(1.0)
        assert mps.bond_sizes() == [2, 2, 2, 2]
        assert qu.fidelity(psi, mps.to_dense()) == pytest.approx(1.0)

    def test_computational_state(self):
        mps = qtn.MPS_computational_state("01+-")
        assert_allclose(
            mps.to_dense(), qu.up() & qu.down() & qu.plus() & qu.minus()
        )


class TestMatrixProductOperatorSpecifics:
    @pytest.mark.parametrize("L", [1, 5])
    def test_MPO_product_operator(self, L):
        psis = [qu.rand_ket(2) for _ in range(L)]
        ops = [qu.rand_matrix(2) for _ in range(L)]
        psif = qu.kron(*ops) @ qu.kron(*psis)
        mps = qtn.MPS_product_state(psis)
        mpo = qtn.MPO_product_operator(ops)
        assert mpo.bond_sizes() == [1] * (L - 1)
        mpsf = mpo.apply(mps)
        assert_allclose(mpsf.to_dense(), psif)


class TestGenericTN:
    def test_TN_rand_reg(self):
        n = 6
        reg = 3
        D = 2
        tn = qtn.TN_rand_reg(n, reg, D=D)
        assert tn.outer_inds() == ()
        assert tn.max_bond() == D
        assert {t.ndim for t in tn} == {reg}
        ket = qtn.TN_rand_reg(n, reg, D=2, phys_dim=2)
        assert set(ket.outer_inds()) == {f"k{i}" for i in range(n)}
        assert ket.max_bond() == D

    @pytest.mark.parametrize("Lx", [3])
    @pytest.mark.parametrize("Ly", [2, 4])
    @pytest.mark.parametrize("beta", [0.13, 0.44])
    @pytest.mark.parametrize("j", [-1.0, +1.0])
    @pytest.mark.parametrize("h", [0.0, 0.1])
    @pytest.mark.parametrize(
        "cyclic", [False, True, (False, True), (True, False)]
    )
    def test_2D_classical_ising_model(self, Lx, Ly, beta, j, h, cyclic):
        tn = qtn.TN2D_classical_ising_partition_function(
            Lx, Ly, beta=beta, j=j, h=h, cyclic=cyclic
        )
        htn = qtn.HTN2D_classical_ising_partition_function(
            Lx, Ly, beta=beta, j=j, h=h, cyclic=cyclic
        )
        Z1 = tn.contract(all, output_inds=())
        Z2 = htn.contract(all, output_inds=())
        assert Z1 == pytest.approx(Z2)

        if not cyclic:
            # skip cyclic as nx has no multibonds for L=2
            import networkx as nx

            G = nx.lattice.grid_graph((Lx, Ly))
            Z3 = qtn.TN_classical_partition_function_from_edges(
                G.edges, beta=beta, j=j, h=h
            ).contract(all, output_inds=())
            assert Z2 == pytest.approx(Z3)
            Z4 = qtn.HTN_classical_partition_function_from_edges(
                G.edges, beta=beta, j=j, h=h
            ).contract(all, output_inds=())
            assert Z3 == pytest.approx(Z4)

    @pytest.mark.parametrize("Lx", [2])
    @pytest.mark.parametrize("Ly", [3])
    @pytest.mark.parametrize("Lz", [4])
    @pytest.mark.parametrize("beta", [0.13, 1 / 4.5])
    @pytest.mark.parametrize("j", [-1.0, +1.0])
    @pytest.mark.parametrize("h", [0.0, 0.1])
    @pytest.mark.parametrize(
        "cyclic", [False, True, (False, True, False), (True, False, True)]
    )
    def test_3D_classical_ising_model(self, Lx, Ly, Lz, beta, j, h, cyclic):
        tn = qtn.TN3D_classical_ising_partition_function(
            Lx, Ly, Lz, beta=beta, j=j, h=h, cyclic=cyclic
        )
        htn = qtn.HTN3D_classical_ising_partition_function(
            Lx, Ly, Lz, beta=beta, j=j, h=h, cyclic=cyclic
        )
        Z1 = tn.contract(all, output_inds=())
        Z2 = htn.contract(all, output_inds=())
        assert Z1 == pytest.approx(Z2)

        if not cyclic:
            # skip cyclic as nx has no multibonds for L=2
            import networkx as nx

            G = nx.lattice.grid_graph((Lx, Ly, Lz))
            Z3 = qtn.TN_classical_partition_function_from_edges(
                G.edges, beta=beta, j=j, h=h
            ).contract(all, output_inds=())
            assert Z2 == pytest.approx(Z3)
            Z4 = qtn.HTN_classical_partition_function_from_edges(
                G.edges, beta=beta, j=j, h=h
            ).contract(all, output_inds=())
            assert Z3 == pytest.approx(Z4)

    def test_2d_classical_ising_varying_j(self):
        L = 5
        beta = 0.3
        edges = qtn.edges_2d_square(L, L)
        np.random.seed(666)
        js = {edge: np.random.normal() for edge in edges}
        tn = qtn.TN_classical_partition_function_from_edges(
            edges, beta=beta, j=lambda i, j: js[i, j]
        )
        assert tn.dtype == "float64"
        x0 = tn.contract(all, output_inds=())
        tn = qtn.HTN_classical_partition_function_from_edges(
            edges, beta=beta, j=lambda i, j: js[i, j]
        )
        assert tn.dtype == "float64"
        x1 = tn.contract(all, output_inds=())
        tn = qtn.TN2D_classical_ising_partition_function(
            L, L, beta=beta, j=lambda i, j: js[i, j]
        )
        assert tn.dtype == "float64"
        x2 = tn.contract(all, output_inds=())
        tn = qtn.HTN2D_classical_ising_partition_function(
            L, L, beta=beta, j=lambda i, j: js[i, j]
        )
        assert tn.dtype == "float64"
        x3 = tn.contract(all, output_inds=())
        assert x0 == pytest.approx(x1)
        assert x1 == pytest.approx(x2)
        assert x2 == pytest.approx(x3)

    def test_3d_classical_ising_varying_j(self):
        L = 3
        beta = 0.3
        edges = qtn.edges_3d_cubic(L, L, L)
        np.random.seed(666)
        js = {edge: np.random.normal() for edge in edges}
        tn = qtn.TN_classical_partition_function_from_edges(
            edges, beta=beta, j=lambda i, j: js[i, j]
        )
        assert tn.dtype == "float64"
        x0 = tn.contract(all, output_inds=())
        tn = qtn.HTN_classical_partition_function_from_edges(
            edges, beta=beta, j=lambda i, j: js[i, j]
        )
        assert tn.dtype == "float64"
        x1 = tn.contract(all, output_inds=())
        tn = qtn.TN3D_classical_ising_partition_function(
            L, L, L, beta=beta, j=lambda i, j: js[i, j]
        )
        assert tn.dtype == "float64"
        x2 = tn.contract(all, output_inds=())
        tn = qtn.HTN3D_classical_ising_partition_function(
            L, L, L, beta=beta, j=lambda i, j: js[i, j]
        )
        assert tn.dtype == "float64"
        x3 = tn.contract(all, output_inds=())
        assert x0 == pytest.approx(x1)
        assert x1 == pytest.approx(x2)
        assert x2 == pytest.approx(x3)

    def test_tn_dimer_covering(self):
        edges = [(0, 1), (1, 2), (2, 3), (3, 0)]
        tn = qtn.TN_dimer_covering_from_edges(edges, cover_count=1)
        assert tn ^ all == pytest.approx(2.0)
        tn = qtn.TN_dimer_covering_from_edges(edges, cover_count=2)
        assert tn ^ all == pytest.approx(1.0)
        edges = [(0, 1), (1, 2), (2, 0)]
        tn = qtn.TN_dimer_covering_from_edges(edges, cover_count=1)
        assert tn ^ all == pytest.approx(0.0)

    def test_tn2d_fillers(self):
        tn = qtn.TN2D_empty(Lx=2, Ly=2, D=2)
        assert isinstance(tn, qtn.TensorNetwork2D)
        assert (
            qtn.TN2D_rand(Lx=2, Ly=2, D=2, seed=42) ^ all
        ) == pytest.approx(qtn.TN2D_rand(Lx=2, Ly=2, D=2, seed=42) ^ all)
        tn = qtn.TN2D_with_value(1.0, Lx=2, Ly=3, D=4)
        assert tn ^ all == pytest.approx(qu.prod(tn.ind_sizes().values()))

    def test_tn3d_fillers(self):
        tn = qtn.TN3D_empty(Lx=2, Ly=2, Lz=2, D=2)
        assert isinstance(tn, qtn.TensorNetwork3D)
        assert (
            qtn.TN3D_rand(Lx=2, Ly=2, Lz=2, D=2, seed=42) ^ all
        ) == pytest.approx(qtn.TN3D_rand(Lx=2, Ly=2, Lz=2, D=2, seed=42) ^ all)
        tn = qtn.TN3D_with_value(1.0, Lx=2, Ly=3, Lz=2, D=2)
        assert tn ^ all == pytest.approx(qu.prod(tn.ind_sizes().values()))


@pytest.mark.parametrize("cyclic", [False, True, (False, True), (True, False)])
def test_tn2d_classical_ising_partition_function(cyclic):
    Lx = 4
    Ly = 5
    coupling = {
        (cooa, coob): float(qu.randn())
        for cooa, coob in qtn.gen_2d_bonds(Lx, Ly, cyclic=cyclic)
    }
    h = qu.randn()
    tn = qtn.TN2D_classical_ising_partition_function(
        Lx,
        Ly,
        beta=0.44,
        j=coupling,
        h=h,
        outputs=[(1, 2), (3, 4)],
        cyclic=cyclic,
    )
    assert tn.outer_inds() == ("s1,2", "s3,4")
    htn = qtn.HTN2D_classical_ising_partition_function(
        Lx,
        Ly,
        beta=0.44,
        j=coupling,
        h=h,
        cyclic=cyclic,
    )
    assert htn.num_indices == Lx * Ly

    if not isinstance(cyclic, tuple):
        cyclic = (cyclic, cyclic)

    assert (tn.is_cyclic_x(), tn.is_cyclic_y()) == cyclic

    assert_allclose(
        tn.contract().data,
        htn.contract(output_inds=("s1,2", "s3,4")).data,
    )


@pytest.mark.parametrize("cyclic", [False, (0, 1, 1), (0, 0, 1)])
def test_tn3d_classical_ising_partition_function(cyclic):
    Lx, Ly, Lz = 2, 3, 3
    coupling = {
        (cooa, coob): float(qu.randn())
        for cooa, coob in qtn.gen_3d_bonds(Lx, Ly, Lz, cyclic=cyclic)
    }
    h = qu.randn()
    tn = qtn.TN3D_classical_ising_partition_function(
        Lx,
        Ly,
        Lz,
        beta=0.44,
        j=lambda cooa, coob: coupling[(cooa, coob)],
        h=h,
        outputs=[(1, 0, 2), (0, 2, 1)],
        cyclic=cyclic,
    )
    assert tn.outer_inds() == ("s0,2,1", "s1,0,2")
    htn = qtn.HTN3D_classical_ising_partition_function(
        Lx,
        Ly,
        Lz,
        beta=0.44,
        j=lambda cooa, coob: coupling[(cooa, coob)],
        h=h,
        cyclic=cyclic,
    )
    assert htn.num_indices == Lx * Ly * Lz

    if not isinstance(cyclic, tuple):
        cyclic = (cyclic, cyclic, cyclic)

    assert (tn.is_cyclic_x(), tn.is_cyclic_y(), tn.is_cyclic_z()) == cyclic

    assert_allclose(
        tn.contract().data,
        htn.contract(output_inds=("s0,2,1", "s1,0,2")).data,
    )


@pytest.mark.parametrize("sites_location", ["side", "diag"])
@pytest.mark.parametrize("outputs", [(), 2, (1, 3)])
def test_all_to_all_classical_partition_functions(sites_location, outputs):
    import numpy as np

    N = 5
    rng = np.random.default_rng(42)
    Jij = {(i, j): rng.normal() for i in range(N) for j in range(i + 1, N)}
    htn = qtn.HTN_classical_partition_function_from_edges(
        edges=Jij.keys(),
        beta=0.179,
        j=Jij,
    )
    Zex = htn.contract(all, output_inds=())

    tn = qtn.TN2D_embedded_classical_ising_partition_function(
        Jij,
        beta=0.179,
        sites_location=sites_location,
        outputs=outputs,
    )

    sites = tuple(tn.gen_sites_present())
    assert len(sites) == N * (N - 1) // 2
    for i, j in sites:
        assert i > j

    if isinstance(outputs, tuple):
        assert set(tn.outer_inds()) == {f"s{i}" for i in outputs}
    else:
        assert tn.outer_inds() == (f"s{outputs}",)
        (t,) = tn._inds_get(f"s{outputs}")
        if sites_location == "side":
            assert "I2,0" in t.tags
        else:
            assert "I2,1" in t.tags
    assert tn.contract(output_inds=()) == pytest.approx(Zex)


def test_tn2d_rand_symm():
    import numpy as np

    tn = qtn.TN2D_rand_symmetric(3, 4, 3)
    ghash = tn.geometry_hash(strict_index_order=True)
    Zex = tn.contract(all, output_inds=())
    rng = np.random.default_rng(42)
    for t in tn:
        t.modify(inds=rng.permutation(t.inds))
    assert tn.geometry_hash(strict_index_order=True) != ghash
    assert tn.contract(all, output_inds=()) == pytest.approx(Zex)



"""Belief propagation for standard tensor networks. This:

- assumes no hyper indices, only standard bonds.
- assumes a single ('dense') tensor per site
- works directly on the '1-norm' i.e. scalar tensor network

This is the simplest version of belief propagation, and is useful for
simple investigations.
"""

import autoray as ar

from quimb.tensor import Tensor, TensorNetwork, rand_uuid
from quimb.tensor.contraction import array_contract
from quimb.utils import oset

from .bp_common import (
    BeliefPropagationCommon,
    combine_local_contractions,
    normalize_message_pair,
    process_loop_series_expansion_weights,
)
from .hd1bp import (
    compute_all_tensor_messages_tree,
)


def initialize_messages(tn, fill_fn=None):
    messages = {}
    for ix, tids in tn.ind_map.items():
        if len(tids) != 2:
            continue
        tida, tidb = tids

        for tid_from, tid_to in [(tida, tidb), (tidb, tida)]:
            t_from = tn.tensor_map[tid_from]
            if fill_fn is not None:
                d = t_from.ind_size(ix)
                m = fill_fn((d,))
            else:
                m = array_contract(
                    arrays=(t_from.data,),
                    inputs=(tuple(range(t_from.ndim)),),
                    output=(t_from.inds.index(ix),),
                )
            messages[ix, tid_to] = m

    return messages


class D1BP(BeliefPropagationCommon):
    """Dense (as in one tensor per site) 1-norm (as in for 'classical' systems)
    belief propagation algorithm. Allows message reuse. This version assumes no
    hyper indices (i.e. a standard tensor network). This is the simplest
    version of belief propagation.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to run BP on.
    messages : dict[(str, int), array_like], optional
        The initial messages to use, effectively defaults to all ones if not
        specified.
    damping : float or callable, optional
        The damping factor to apply to messages. This simply mixes some part
        of the old message into the new one, with the final message being
        ``damping * old + (1 - damping) * new``. This makes convergence more
        reliable but slower.
    update : {'sequential', 'parallel'}, optional
        Whether to update messages sequentially (newly computed messages are
        immediately used for other updates in the same iteration round) or in
        parallel (all messages are comptued using messages from the previous
        round only). Sequential generally helps convergence but parallel can
        possibly converge to differnt solutions.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    local_convergence : bool, optional
        Whether to allow messages to locally converge - i.e. if all their
        input messages have converged then stop updating them.
    contract_every : int, optional
        If not None, 'contract' (via BP) the tensor network every
        ``contract_every`` iterations. The resulting values are stored in
        ``zvals`` at corresponding points ``zval_its``.
    inplace : bool, optional
        Whether to perform any operations inplace on the input tensor network.

    Attributes
    ----------
    tn : TensorNetwork
        The target tensor network.
    messages : dict[(str, int), array_like]
        The current messages. The key is a tuple of the index and tensor id
        that the message is being sent to.
    key_pairs : dict[(str, int), (str, int)]
        A dictionary mapping the key of a message to the key of the message
        propagating in the opposite direction.
    """

    def __init__(
        self,
        tn: TensorNetwork,
        *,
        messages=None,
        damping=0.0,
        update="sequential",
        normalize=None,
        distance=None,
        local_convergence=True,
        message_init_function=None,
        contract_every=None,
        inplace=False,
    ):
        super().__init__(
            tn=tn,
            damping=damping,
            update=update,
            normalize=normalize,
            distance=distance,
            contract_every=contract_every,
            inplace=inplace,
        )

        self.local_convergence = local_convergence

        if callable(messages):
            self.messages = initialize_messages(self.tn, messages)
        elif messages is None:
            self.messages = initialize_messages(self.tn, message_init_function)
        else:
            self.messages = messages

        # record which messages touch which tids, for efficient updates
        self.touched = oset()
        self.key_pairs = {}
        for ix, tids in tn.ind_map.items():
            if len(tids) != 2:
                continue
            tida, tidb = tids
            self.key_pairs[ix, tidb] = (ix, tida)
            self.key_pairs[ix, tida] = (ix, tidb)

    def iterate(self, tol=5e-6):
        if (not self.local_convergence) or (not self.touched):
            # assume if asked to iterate that we want to check all messages
            self.touched = oset(self.tn.tensor_map)

        ncheck = len(self.touched)
        nconv = 0
        max_mdiff = -1.0
        new_touched = oset()

        def _compute_ms(tid):
            t = self.tn.tensor_map[tid]
            new_ms = compute_all_tensor_messages_tree(
                t.data,
                [self.messages[ix, tid] for ix in t.inds],
                self.backend,
            )
            new_ms = [self._normalize_fn(m) for m in new_ms]
            new_ks = [self.key_pairs[ix, tid] for ix in t.inds]

            return new_ks, new_ms

        def _update_m(key, new_m):
            nonlocal nconv, max_mdiff

            old_m = self.messages[key]

            # pre-damp distance
            mdiff = self._distance_fn(old_m, new_m)

            if self.damping:
                new_m = self._damping_fn(old_m, new_m)

            # # post-damp distance
            # mdiff = self._distance_fn(old_m, new_m)

            if mdiff > tol:
                # mark distination tid for update
                new_touched.add(key[1])
            else:
                nconv += 1

            max_mdiff = max(max_mdiff, mdiff)
            self.messages[key] = new_m

        if self.update == "sequential":
            # compute each new message and immediately re-insert it
            while self.touched:
                tid = self.touched.pop()
                keys, new_ms = _compute_ms(tid)
                for key, new_m in zip(keys, new_ms):
                    _update_m(key, new_m)

        elif self.update == "parallel":
            new_data = {}
            # compute all new messages
            while self.touched:
                tid = self.touched.pop()
                keys, new_ms = _compute_ms(tid)
                for key, new_m in zip(keys, new_ms):
                    new_data[key] = new_m
            # insert all new messages
            for key, new_m in new_data.items():
                _update_m(key, new_m)

        self.touched = new_touched

        return {
            "nconv": nconv,
            "ncheck": ncheck,
            "max_mdiff": max_mdiff,
        }

    def normalize_message_pairs(self):
        """Normalize all messages such that for each bond `<m_i|m_j> = 1` and
        `<m_i|m_i> = <m_j|m_j>` (but in general != 1).
        """
        for ix, tids in self.tn.ind_map.items():
            if len(tids) != 2:
                continue
            tida, tidb = tids
            mi = self.messages[ix, tida]
            mj = self.messages[ix, tidb]
            mi, mj = normalize_message_pair(mi, mj)
            self.messages[ix, tida] = mi
            self.messages[ix, tidb] = mj

    def normalize_tensors(self, strip_exponent=True):
        """Normalize every local tensor contraction so that it equals 1. Gather
        the overall normalization factor into ``self.exponent`` and the sign
        into ``self.sign`` by default.

        Parameters
        ----------
        strip_exponent : bool, optional
            Whether to collect the sign and exponent. If ``False`` then the
            value of the BP contraction is set to 1.
        """
        for tid, t in self.tn.tensor_map.items():
            tval = self.local_tensor_contract(tid)
            tabs = ar.do("abs", tval)
            tsgn = tval / tabs
            tlog = ar.do("log10", tabs)
            t /= tsgn * tabs
            if strip_exponent:
                self.sign = tsgn * self.sign
                self.exponent = tlog + self.exponent

    def get_gauged_tn(self):
        """Gauge the original TN by inserting the BP-approximated transfer
        matrix eigenvectors, which may be complex. The BP-contraction of this
        gauged network is then simply the product of zeroth entries of each
        tensor.
        """
        tng = self.tn.copy()
        for ind, tids in self.tn.ind_map.items():
            tida, tidb = tids
            ka = (ind, tida)
            kb = (ind, tidb)
            ma = self.messages[ka]
            mb = self.messages[kb]

            el, ev = ar.do("linalg.eig", ar.do("outer", ma, mb))
            k = ar.do("argsort", -ar.do("abs", el))
            ev = ev[:, k]
            Uinv = ev
            U = ar.do("linalg.inv", ev)
            tng._insert_gauge_tids(U, tida, tidb, Uinv)
        return tng

    def get_cluster(self, tids):
        """Get the region of tensors given by `tids`, with the messages
        on the border contracted in, removing those dangling indices.

        Parameters
        ----------
        tids : sequence of int
            The tensor ids forming a region.

        Returns
        -------
        TensorNetwork
        """
        # take copy as we are going contract messages in
        tnr = self.tn._select_tids(tids, virtual=False)
        oixr = tnr.outer_inds()
        for ix in oixr:
            # get the tensor this index belongs to
            (tid,) = tnr._get_tids_from_inds(ix)
            t = tnr.tensor_map[tid]
            # contract the message in, removing index
            t.vector_reduce_(ix, self.messages[ix, tid])
        return tnr

    def get_cluster_excited(self, tids):
        """Get the local tensor network for ``tids`` with BP messages inserted
        on the boundary and excitation projectors inserted on the inner bonds.
        See https://arxiv.org/abs/2409.03108 for more details.
        """
        stn = self.tn._select_tids(tids, virtual=False)

        for ix, tids in tuple(stn.ind_map.items()):
            if ix in stn._inner_inds:
                # insert inner excitation projector
                tidl, tidr = tids
                ml = self.messages[ix, tidl]
                mr = self.messages[ix, tidr]
                # form outer product
                p0 = ar.do("einsum", "i,j->ij", ml, mr)
                # subtract from identity
                pe = ar.do("eye", ar.do("shape", p0)[0]) - p0
                # absorb into one of tensors
                stn.tensor_map[tidr].gate_(pe, ix)
            else:
                # insert boundary message
                (tid,) = tids
                m = self.messages[ix, tid]
                t = stn.tensor_map[tid]
                t.vector_reduce_(ix, m)

        return stn

    def contract_loop_series_expansion(
        self,
        gloops=None,
        multi_excitation_correct=True,
        tol_correction=1e-12,
        maxiter_correction=100,
        strip_exponent=False,
        optimize="auto-hq",
        **contract_opts,
    ):
        """Contract the tensor network using the same procedure as
        in https://arxiv.org/abs/2409.03108 - "Loop Series Expansions for
        Tensor Networks".

        Parameters
        ----------
        gloops : int or iterable of tuples, optional
            The gloop sizes to use. If an integer, then generate all gloop
            sizes up to this size. If a tuple, then use these gloops.
        multi_excitation_correct : bool, optional
            Whether to use the multi-excitation correction. If ``True``, then
            the free energy is refined iteratively until self consistent.
        tol_correction : float, optional
            The tolerance for the multi-excitation correction.
        maxiter_correction : int, optional
            The maximum number of iterations for the multi-excitation
            correction.
        strip_exponent : bool, optional
            Whether to strip the exponent from the final result. If ``True``
            then the returned result is ``(mantissa, exponent)``.
        optimize : str or PathOptimizer, optional
            The path optimizer to use when contracting the messages.
        contract_opts
            Other options supplied to ``TensorNetwork.contract``.
        """
        self.normalize_message_pairs()
        # accrues BP estimate into self.sign and self.exponent
        self.normalize_tensors()

        if isinstance(gloops, int):
            gloops = tuple(self.tn.gen_gloops(max_size=gloops))
        else:
            gloops = tuple(gloops)

        weights = {}
        for gloop in gloops:
            # get local tensor network with boundary
            # messages and exctiation projectors
            etn = self.get_cluster_excited(gloop)
            # contract it to get local weight!
            weights[tuple(gloop)] = etn.contract(
                optimize=optimize, **contract_opts
            )

        return process_loop_series_expansion_weights(
            weights,
            mantissa=self.sign,
            exponent=self.exponent,
            multi_excitation_correct=multi_excitation_correct,
            tol_correction=tol_correction,
            maxiter_correction=maxiter_correction,
            strip_exponent=strip_exponent,
        )

    def local_tensor_contract(self, tid):
        """Contract the messages around tensor ``tid``."""
        t = self.tn.tensor_map[tid]
        arrays = [t.data]
        inputs = [tuple(range(t.ndim))]
        for i, ix in enumerate(t.inds):
            m = self.messages[ix, tid]
            arrays.append(m)
            inputs.append((i,))

        return array_contract(
            arrays=arrays,
            inputs=inputs,
            output=(),
        )

    def local_message_contract(self, ix):
        """Contract the messages at index ``ix``."""
        tids = self.tn.ind_map[ix]
        if len(tids) != 2:
            return None
        tida, tidb = tids
        return self.messages[ix, tida] @ self.messages[ix, tidb]

    def contract(
        self,
        strip_exponent=False,
        check_zero=True,
        **kwargs,
    ):
        """Estimate the contraction of the tensor network."""

        zvals = [
            (self.local_tensor_contract(tid), 1) for tid in self.tn.tensor_map
        ] + [(self.local_message_contract(ix), -1) for ix in self.tn.ind_map]

        return combine_local_contractions(
            zvals,
            backend=self.backend,
            strip_exponent=strip_exponent,
            check_zero=check_zero,
            mantissa=self.sign,
            exponent=self.exponent,
            **kwargs,
        )

    def contract_with_loops(
        self,
        max_loop_length=None,
        min_loop_length=1,
        optimize="auto-hq",
        strip_exponent=False,
        check_zero=True,
        **contract_opts,
    ):
        """Estimate the contraction of the tensor network, including loop
        corrections.
        """
        self.normalize_message_pairs()
        self.normalize_tensors()

        zvals = []

        for loop in self.tn.gen_paths_loops(max_loop_length=max_loop_length):
            if len(loop) < min_loop_length:
                continue

            # get the loop local patch
            ltn = self.tn.select_path(loop)

            # attach boundary messages

            for ix, tids in tuple(ltn.ind_map.items()):
                if ix in loop:
                    continue

                elif len(tids) == 1:
                    # outer index -> cap it with messages
                    (tid,) = tids
                    ltn |= Tensor(self.messages[ix, tid], [ix])

                else:
                    # non-loop internal index -> cut it with messages
                    tida, tidb = tids
                    ma = self.messages[ix, tida]
                    mb = self.messages[ix, tidb]
                    lix = rand_uuid()
                    rix = rand_uuid()
                    ltn._cut_between_tids(tida, tidb, lix, rix)
                    ltn |= Tensor(ma, [lix])
                    ltn |= Tensor(mb, [rix])

            zvals.append((ltn.contract(optimize=optimize, **contract_opts), 1))

        return combine_local_contractions(
            zvals,
            backend=self.backend,
            strip_exponent=strip_exponent,
            check_zero=check_zero,
            mantissa=self.sign,
            exponent=self.exponent,
        )

    def contract_gloop_expand(
        self,
        gloops=None,
        autocomplete=True,
        strip_exponent=False,
        check_zero=True,
        optimize="auto-hq",
        combine="prod",
        **contract_opts,
    ):
        """Contract the tensor network using generalized loop cluster
        expansion.
        """
        from .regions import gen_region_counts

        if isinstance(gloops, int):
            max_size = gloops
            gloops = None
        else:
            max_size = None

        if gloops is None:
            gloops = tuple(self.tn.gen_gloops(max_size=max_size))
        else:
            gloops = tuple(gloops)

        if combine == "sum":
            # make sure each contraction has the same BP-scaled environment
            self.normalize_message_pairs()
            self.normalize_tensors()

        zvals = []
        for r, c in gen_region_counts(gloops, autocomplete=autocomplete):
            # XXX: autoreduce intersecting clusters to gloops?
            tnr = self.get_cluster(r)
            zr = tnr.contract(optimize=optimize, **contract_opts)

            zvals.append((zr, c))

        if combine == "sum":
            mantissa = self.sign * sum(zr * cr for zr, cr in zvals)
            if strip_exponent:
                return mantissa, self.exponent
            return mantissa * 10**self.exponent

        return combine_local_contractions(
            zvals,
            backend=self.backend,
            strip_exponent=strip_exponent,
            check_zero=check_zero,
            mantissa=self.sign,
            exponent=self.exponent,
        )


def contract_d1bp(
    tn,
    *,
    max_iterations=1000,
    tol=5e-6,
    damping=0.0,
    diis=False,
    update="sequential",
    normalize=None,
    distance=None,
    tol_abs=None,
    tol_rolling_diff=None,
    local_convergence=True,
    strip_exponent=False,
    check_zero=True,
    info=None,
    progbar=False,
    **contract_opts,
):
    """Estimate the contraction of standard tensor network ``tn`` using dense
    1-norm belief propagation.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to contract, it should have no dangling or hyper
        indices.
    max_iterations : int, optional
        The maximum number of iterations to perform.
    tol : float, optional
        The convergence tolerance for messages.
    damping : float, optional
        The damping parameter to use, defaults to no damping.
    diis : bool or dict, optional
        Whether to use direct inversion in the iterative subspace to
        help converge the messages by extrapolating to low error guesses.
        If a dict, should contain options for the DIIS algorithm. The
        relevant options are {`max_history`, `beta`, `rcond`}.
    update : {'sequential', 'parallel'}, optional
        Whether to update messages sequentially or in parallel.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    tol_abs : float, optional
        The absolute convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``.
    tol_rolling_diff : float, optional
        The rolling mean convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``. This is used to stop
        running when the messages are just bouncing around the same level,
        without any overall upward or downward trends, roughly speaking.
    local_convergence : bool, optional
        Whether to allow messages to locally converge - i.e. if all their
        input messages have converged then stop updating them.
    strip_exponent : bool, optional
        Whether to return the mantissa and exponent separately.
    check_zero : bool, optional
        Whether to check for zero values and return zero early.
    info : dict, optional
        If supplied, the following information will be added to it:
        ``converged`` (bool), ``iterations`` (int), ``max_mdiff`` (float),
        ``rolling_abs_mean_diff`` (float).
    progbar : bool, optional
        Whether to show a progress bar.
    """
    bp = D1BP(
        tn,
        damping=damping,
        local_convergence=local_convergence,
        update=update,
        normalize=normalize,
        distance=distance,
        **contract_opts,
    )
    bp.run(
        max_iterations=max_iterations,
        diis=diis,
        tol=tol,
        tol_abs=tol_abs,
        tol_rolling_diff=tol_rolling_diff,
        info=info,
        progbar=progbar,
    )
    return bp.contract(
        strip_exponent=strip_exponent,
        check_zero=check_zero,
    )



import pytest

import quimb.tensor as qtn
import quimb.tensor.belief_propagation as qbp


@pytest.mark.parametrize("dtype", ["float32", "complex64"])
def test_contract_tree_exact(dtype):
    psi = qtn.TN_rand_tree(20, 3, 2, dtype=dtype)
    norm2 = psi.H @ psi
    info = {}
    norm2_bp = qbp.contract_l2bp(psi, info=info, progbar=True)
    assert info["converged"]
    assert norm2_bp == pytest.approx(norm2, rel=5e-5)


@pytest.mark.parametrize("dtype", ["float32", "complex64"])
def test_contract_loopy_approx(dtype):
    peps = qtn.PEPS.rand(3, 4, 3, dtype=dtype, seed=42)
    norm_ex = peps.H @ peps
    info = {}
    norm_bp = qbp.contract_l2bp(peps, damping=0.1, info=info, progbar=True)
    assert info["converged"]
    assert norm_bp == pytest.approx(norm_ex, rel=0.2)


@pytest.mark.parametrize("damping", [0.0, 0.1])
@pytest.mark.parametrize("dtype", ["float32", "complex64"])
def test_compress_loopy(damping, dtype):
    peps = qtn.PEPS.rand(3, 4, 3, seed=42, dtype=dtype)
    # test that using the BP compression gives better fidelity than purely
    # local, naive compression scheme
    peps_c1 = peps.compress_all(max_bond=2)
    info = {}
    peps_c2 = qbp.compress_l2bp(
        peps, max_bond=2, damping=damping, info=info, progbar=True
    )
    assert info["converged"]
    fid1 = peps_c1.H @ peps_c2
    fid2 = peps_c2.H @ peps_c2
    assert abs(fid2) > abs(fid1)


@pytest.mark.parametrize("dtype", ["float32", "complex64"])
def test_contract_double_layer_tree_exact(dtype):
    # generate a random binary tree
    edges = qtn.edges_tree_rand(10, max_degree=3, seed=42)
    # generate a random tree product state and operator on this tree
    tps = qtn.TN_from_edges_rand(
        edges,
        3,
        phys_dim=2,
        site_ind_id="k{}",
        dtype=dtype,
        seed=111,
    )
    tpo = qtn.TN_from_edges_rand(
        edges,
        3,
        phys_dim=2,
        site_ind_id=("k{}", "b{}"),
        dtype=dtype,
        seed=222,
    )
    # join into double layer tree
    tn = qtn.tensor_network_apply_op_vec(tpo, tps, contract=False)
    assert tn.num_tensors == 20

    norm_ex = tn.H @ tn
    info = {}
    norm_bp = qbp.contract_l2bp(tn, info=info, progbar=True)
    assert info["converged"]

    assert norm_bp == pytest.approx(norm_ex, rel=5e-6)


@pytest.mark.parametrize("dtype", ["float32", "complex64"])
@pytest.mark.parametrize("damping", [0.0, 0.1])
@pytest.mark.parametrize("update", ["parallel", "sequential"])
def test_compress_double_layer_loopy(dtype, damping, update):
    peps = qtn.PEPS.rand(3, 4, bond_dim=3, seed=42, dtype=dtype)
    pepo = qtn.PEPO.rand(3, 4, bond_dim=2, seed=42, dtype=dtype)

    tn_lazy = qtn.tensor_network_apply_op_vec(pepo, peps, contract=False)
    assert tn_lazy.num_tensors == 24

    # compress using basic local compression
    tn_eager = qtn.tensor_network_apply_op_vec(pepo, peps, contract=True)
    assert tn_eager.num_tensors == 12
    tn_eager.compress_all_(max_bond=3, canonize=False)
    fid_basic = abs(tn_eager.H @ tn_lazy)

    # compress using BP
    info = {}
    tn_bp = qbp.compress_l2bp(
        tn_lazy,
        max_bond=3,
        damping=damping,
        update=update,
        info=info,
        progbar=True,
    )
    assert info["converged"]
    assert tn_bp.num_tensors == 12

    # assert we did better than basic local compression
    fid_bp = abs(tn_bp.H @ tn_lazy)
    assert fid_bp > fid_basic



import math

import pytest

import quimb.operator as qop


def test_basic_int():
    hs = qop.HilbertSpace(10)
    print(hs)
    assert hs.nsites == 10
    assert hs.size == 1024
    assert len(hs.sites) == 10
    for i in range(10):
        assert hs.site_to_reg(i) == i
        assert hs.reg_to_site(i) == i

    fc = hs.rand_flatconfig()
    assert len(fc) == 10
    for xi, d in zip(fc, hs.sizes):
        assert 0 <= xi < d


def test_basic_sequence():
    hs = qop.HilbertSpace(["a", "b", "c"], dims=3)
    print(hs)
    assert hs.nsites == 3
    assert hs.size == 27
    assert len(hs.sites) == 3
    for i, s in enumerate(["a", "b", "c"]):
        assert hs.site_to_reg(s) == i
        assert hs.reg_to_site(i) == s

    fc = hs.rand_flatconfig()
    assert len(fc) == 3
    for xi, d in zip(fc, hs.sizes):
        assert 0 <= xi < d


def test_basic_mapping():
    hs = qop.HilbertSpace({"Z": 2, "X": 3, "Y": 4}, order=False)
    print(hs)
    assert hs.nsites == 3
    assert hs.size == 24
    assert len(hs.sites) == 3
    assert hs.site_to_reg("Z") == 0
    assert hs.site_size("Z") == 2
    assert hs.site_to_reg("X") == 1
    assert hs.site_size("X") == 3
    assert hs.site_to_reg("Y") == 2
    assert hs.site_size("Y") == 4
    assert hs.reg_to_site(0) == "Z"
    assert hs.reg_to_site(1) == "X"
    assert hs.reg_to_site(2) == "Y"
    assert hs.sites == ("Z", "X", "Y")
    assert tuple(map(int, hs.sizes)) == (2, 3, 4)
    assert tuple(map(int, hs.strides)) == (12, 4, 1)


def test_basic_mapping_sorted():
    hs = qop.HilbertSpace({"Z": 2, "X": 3, "Y": 4}, order=True)
    print(hs)
    assert hs.nsites == 3
    assert hs.size == 2 * 3 * 4
    assert len(hs.sites) == 3
    assert hs.site_to_reg("X") == 0
    assert hs.site_size("X") == 3
    assert hs.site_to_reg("Y") == 1
    assert hs.site_size("Y") == 4
    assert hs.site_to_reg("Z") == 2
    assert hs.site_size("Z") == 2
    assert hs.reg_to_site(0) == "X"
    assert hs.reg_to_site(1) == "Y"
    assert hs.reg_to_site(2) == "Z"
    assert hs.sites == ("X", "Y", "Z")
    assert tuple(map(int, hs.sizes)) == (3, 4, 2)
    assert tuple(map(int, hs.strides)) == (8, 2, 1)


def test_mixed_radix_sampling():
    hs = qop.HilbertSpace({"Z": 2, "X": 3, "Y": 4}, order=True)
    for _ in range(100):
        config = hs.rand_config()
        assert set(config.keys()) == {"X", "Y", "Z"}
        assert 0 <= config["X"] < 3
        assert 0 <= config["Y"] < 4
        assert 0 <= config["Z"] < 2
        rank = hs.config_to_rank(config)
        assert 0 <= rank < hs.size


@pytest.mark.parametrize("sector", [0, 1])
def test_basic_z2_symmetry_sampling(sector):
    hs = qop.HilbertSpace(6, sector=sector, symmetry="Z2")
    print(hs)
    assert hs.size == 2 ** (hs.nsites - 1)
    for _ in range(100):
        config = hs.rand_config()
        assert sum(config.values()) % 2 == sector
        rank = hs.config_to_rank(config)
        assert 0 <= rank < hs.size


@pytest.mark.parametrize("sector", [0, 1, 2, 3, 4, 5, 6])
def test_basic_u1_symmetry_sampling(sector):
    hs = qop.HilbertSpace(6, symmetry="U1", sector=sector)
    print(hs)
    assert hs.size == math.comb(6, sector)
    for _ in range(100):
        config = hs.rand_config()
        assert sum(config.values()) == sector
        rank = hs.config_to_rank(config)
        assert 0 <= rank < hs.size


@pytest.mark.parametrize("sectora", [0, 1, 2, 3, 4, 5, 6])
@pytest.mark.parametrize("sectorb", [0, 1, 2, 3, 4])
def test_basic_u1u1_symmetry_sampling(sectora, sectorb):
    hs = qop.HilbertSpace(
        10, symmetry="U1U1", sector=((6, sectora), (4, sectorb))
    )
    print(hs)
    assert hs.size == math.comb(6, sectora) * math.comb(4, sectorb)
    for _ in range(10):
        config = hs.rand_config()
        suma = sum(config[i] for i in range(6))
        sumb = sum(config[i] for i in range(6, 10))
        assert suma == sectora
        assert sumb == sectorb
        rank = hs.config_to_rank(config)
        assert 0 <= rank < hs.size



import functools
import math

# a style to use for matplotlib that works with light and dark backgrounds
NEUTRAL_STYLE = {
    "axes.edgecolor": (0.5, 0.5, 0.5),
    "axes.facecolor": (0, 0, 0, 0),
    "axes.grid": True,
    "axes.labelcolor": (0.5, 0.5, 0.5),
    "axes.spines.right": False,
    "axes.spines.top": False,
    "figure.facecolor": (0, 0, 0, 0),
    "grid.alpha": 0.1,
    "grid.color": (0.5, 0.5, 0.5),
    "legend.frameon": False,
    "text.color": (0.5, 0.5, 0.5),
    "xtick.color": (0.5, 0.5, 0.5),
    "xtick.minor.visible": True,
    "ytick.color": (0.5, 0.5, 0.5),
    "ytick.minor.visible": True,
}


def default_to_neutral_style(fn):
    """Wrap a function or method to use the neutral style by default."""

    @functools.wraps(fn)
    def wrapper(
        *args,
        style="neutral",
        show_and_close=True,
        clear_previous=False,
        **kwargs,
    ):
        import matplotlib.pyplot as plt

        if clear_previous:
            from IPython import display

            # clear old plots
            display.clear_output(wait=True)

        if style == "neutral":
            style = NEUTRAL_STYLE
        elif not style:
            style = {}

        with plt.style.context(style):
            out = fn(*args, **kwargs)

            if show_and_close:
                plt.show()
                plt.close()

            return out

    return wrapper


def _ensure_dict(k, v):
    import numpy as np
    from .schematic import hash_to_color, get_color

    # ensure is a dictionaty
    if not isinstance(v, dict):
        v = {"y": v}
    v["y"] = np.array(v["y"])

    if v["y"].size < 2:
        return None

    # make sure x-coords exists explicitly
    if "x" not in v:
        v["x"] = np.arange(v["y"].size)
    else:
        v["x"] = np.array(v["x"])

    # set label as data name by default
    v.setdefault("label", k)

    if v.get("color", None) is None:
        label = v["label"]
        if label is None:
            v["color"] = get_color("blue")
        else:
            v["color"] = hash_to_color(k, vmin=0.75, vmax=0.85)

    return v


@default_to_neutral_style
def plot(*args, **kwargs):
    """Wrapper around `matplotlib.pyplot.plot` to use the neutral style by
    default (`style="neutral") as well as showing and closing the plot by
    default (`show_and_close=True`).
    """
    import matplotlib.pyplot as plt

    return plt.plot(*args, **kwargs)


@default_to_neutral_style
def plot_multi_series_zoom(
    data,
    zoom="auto",
    zoom_max=100,
    zoom_marker="|",
    zoom_markersize=3,
    xlabel="Iteration",
    width=8,
    height=None,
    figsize=None,
    **kwargs,
):
    """Plot possibly multiple series of data, using the asinh scale for an
    overview and a linear scale for a zoomed in final section.

    Parameters
    ----------
    data : dict[dict], dict[array], dict, array, optional
        The data to plot.
    """
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    if isinstance(data, dict) and "y" not in data:
        # multiple plain, or configured, sequences supplied
        data = [_ensure_dict(k, v) for k, v in data.items()]
    else:
        # single plain, or configured, sequence supplied
        data = [_ensure_dict(None, data)]

    # remove any empty data
    data = [d for d in data if d is not None]

    nrows = len(data)

    if nrows == 0:
        return None, None

    if figsize is None:
        if height is None:
            height = 2 * nrows**0.8
        figsize = (width, height)

    fig, axs = plt.subplots(
        nrows=nrows,
        ncols=2,
        figsize=figsize,
        width_ratios=(3, 2),
        gridspec_kw={"wspace": 0.05, "hspace": 0.10},
        squeeze=False,
    )

    n = max(d["x"][-1] for d in data)
    if zoom is not None:
        if zoom == "auto":
            zoom = min(zoom_max, n // 2)
    nz = n - zoom

    for i, d in enumerate(data):
        # get data and correct zoomed range
        x = d.pop("x")
        y = d.pop("y")
        iz = min(range(x.size), key=lambda i: x[i] < nz)

        label = d.pop("label")
        color = d.pop("color")
        yscale = d.pop("yscale", kwargs.get("yscale", "linear"))

        # plot overview
        ax = axs[i, 0]
        ax.plot(
            x,
            y,
            color=color,
            linewidth=1,
        )
        if label is not None:
            ax.text(
                0.05,
                1.0,
                label,
                color=color,
                transform=ax.transAxes,
                ha="left",
                va="top",
            )
        # x props
        ax.set_xscale("asinh", linear_width=20)
        ax.xaxis.set_major_locator(
            mpl.ticker.AsinhLocator(20, numticks=6, subs=range(10))
        )
        # y props
        ax.tick_params(axis="y", colors=color, which="both")
        if yscale == "linear":
            ax.yaxis.set_major_formatter(
                mpl.ticker.ScalarFormatter(useOffset=False)
            )
        elif yscale == "asinh":
            ax.yaxis.set_major_locator(mpl.ticker.AsinhLocator(1, numticks=6))
        else:
            ax.set_yscale(yscale)

        # highlight zoomed range
        ax.axvspan(nz, n, alpha=0.15, color=(0.5, 0.5, 0.5))

        # plot zoom
        ax = axs[i, 1]
        ax.plot(
            x[iz:],
            y[iz:],
            color=color,
            marker=zoom_marker,
            markersize=zoom_markersize,
        )
        # y props
        ax.yaxis.tick_right()
        ax.spines["left"].set_visible(False)
        ax.spines["right"].set_visible(True)
        ax.tick_params(axis="y", colors=color, which="both")
        if yscale == "linear":
            ax.yaxis.set_major_formatter(
                mpl.ticker.ScalarFormatter(useOffset=False)
            )
        else:
            ax.set_yscale(yscale)

    # remove ticklabels on all but last row
    for i in range(nrows - 1):
        axs[i, 0].tick_params(axis="x", labelbottom=False)
        axs[i, 1].tick_params(axis="x", labelbottom=False)

    # set x-limits to just cover full range of data
    for i in range(nrows):
        axs[i, 0].set_xlim(0.0 - 0.5, n + 0.5)
        axs[i, 1].set_xlim(nz - 0.5, n + 0.5)

    # make the xticklabels appear like [0, 1, 10, 100, ...]
    axs[-1, 0].xaxis.set_minor_formatter(mpl.ticker.NullFormatter())
    axs[-1, 0].xaxis.set_major_formatter(
        mpl.ticker.FuncFormatter(
            lambda x, _: f"{int(x):,}"
            if math.isclose(x, 0)
            or (x > 1 and math.isclose(math.log10(x) % 1, 0))
            else ""
        )
    )

    # set x-labels
    axs[-1, 0].set_xlabel(f"{xlabel} (full)")
    axs[-1, 1].set_xlabel(f"{xlabel} (zoom)")

    return fig, axs



import itertools
import math

import numpy as np
import pytest
from numpy.testing import assert_allclose

import quimb as qu
import quimb.tensor as qtn


def rand_reg_graph(reg, n, seed=None):
    import networkx as nx

    G = nx.random_regular_graph(reg, n, seed=seed)
    return G


def graph_to_qsim(G, gamma0=-0.743043, beta0=0.754082):
    n = G.number_of_nodes()

    # add all the gates
    circ = f"{n}\n"
    for i in range(n):
        circ += f"H {i}\n"
    for i, j in G.edges:
        circ += f"Rzz {gamma0} {i} {j}\n"
    for i in range(n):
        circ += f"Rx {beta0} {i}\n"

    return circ


def random_a2a_circ(L, depth, seed=42):
    rng = np.random.default_rng(seed)

    qubits = np.arange(L)
    gates = []

    for i in range(L):
        gates.append((0, "h", i))

    for d in range(depth):
        rng.shuffle(qubits)

        for i in range(0, L - 1, 2):
            g = rng.choice(["cx", "cy", "cz", "iswap"])
            gates.append((d, g, qubits[i], qubits[i + 1]))

        for q in qubits:
            g = rng.choice(["rx", "ry", "rz"])
            gates.append((d, g, rng.normal(1.0, 0.5), q))

    circ = qtn.Circuit(L)
    circ.apply_gates(gates)

    return circ


def qft_circ(n, swaps=True, **circuit_opts):
    circ = qtn.Circuit(n, **circuit_opts)

    for i in range(n):
        circ.h(i)
        for j, m in zip(range(i + 1, n), itertools.count(2)):
            circ.cu1(2 * math.pi / 2**m, j, i)

    if swaps:
        for i in range(n // 2):
            circ.swap(i, n - i - 1)

    return circ


def swappy_circ(n, depth):
    circ = qtn.Circuit(n)

    for d in range(depth):
        pairs = np.random.permutation(np.arange(n))

        for i in range(n // 2):
            qi = pairs[2 * i]
            qj = pairs[2 * i + 1]

            gate = np.random.choice(["FSIM", "SWAP"])
            if gate == "FSIM":
                params = np.random.randn(2)
            elif gate == "FSIMG":
                params = np.random.randn(5)
            else:
                params = ()

            circ.apply_gate(gate, *params, qi, qj)

    return circ


def example_openqasm2_qft():
    return """
    // quantum Fourier transform

    OPENQASM 2.0;
    include "qelib1.inc";

    qreg q[4];
    creg c[4];
    x q[0];
    x q[2];
    barrier q;
    h q[0];
    cu1(pi/2) q[1],q[0];
    h q[1];
    cu1(pi/4) q[2],q[0];
    cu1(pi/2) q[2],q[1];
    /*
    This is a multi line comment.
    */
    h q[2];
    cu1(pi/8) q[3],q[0];
    cu1(pi/4) q[3],q[1];
    cu1(pi/2) q[3],q[2];
    h q[3];

    measure q -> c;
    """


class TestCircuit:
    def test_prepare_GHZ(self):
        qc = qtn.Circuit(3)
        gates = [
            ("H", 0),
            ("H", 1),
            ("CNOT", 1, 2),
            ("CNOT", 0, 2),
            ("H", 0),
            ("H", 1),
            ("H", 2),
        ]
        qc.apply_gates(gates)
        assert qu.expec(qc.psi.to_dense(), qu.ghz_state(3)) == pytest.approx(1)
        counts = qc.simulate_counts(1024)
        assert len(counts) == 2
        assert "000" in counts
        assert "111" in counts
        assert counts["000"] + counts["111"] == 1024

    def test_from_qsim(self):
        G = rand_reg_graph(reg=3, n=18, seed=42)
        qsim = graph_to_qsim(G)
        qc = qtn.Circuit.from_qsim_str(qsim)
        assert (qc.psi.H & qc.psi) ^ all == pytest.approx(1.0)

    def test_from_openqasm2(self):
        qc = qtn.Circuit.from_openqasm2_str(example_openqasm2_qft())
        assert (qc.psi.H & qc.psi) ^ all == pytest.approx(1.0)

    def test_openqasm2_custom_gates(self):
        circ = qtn.Circuit.from_openqasm2_str(
            """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[3];

        gate hello a, b {
            h a;
            cx a, b;
            u3(0.1, 0.2, 0.3) b;
        }

        gate world(param1, θ) q
        {
            u2(θ / 2, param1) q;
            u2(param1, θ / 2) q;
        }

        hello q[0], q[1];
        world(0.1, 0.2) q[2];
        hello q[2], q[1];
        """
        )
        assert [g.label for g in circ.gates] == [
            "H",
            "CX",
            "U3",
            "U2",
            "U2",
            "H",
            "CX",
            "U3",
        ]

    def test_openqasm2_custom_nested_gates(self):
        circ = qtn.Circuit.from_openqasm2_str(
            """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[3];

        gate cphase(θ) a, b
        {
            U3(0, 0, θ / 2) a;
            CX a, b;
            U3(0, 0, -θ / 2) b;
            CX a, b;
            U3(0, 0, θ / 2) b;
        }

        gate doublecphase(θ) a, b, c {
            cphase(θ) a, b;
            cphase(θ) b, c;
        }

        doublecphase(0.1) q[0], q[1], q[2];
        doublecphase(0.2) q[2], q[0], q[1];
        """
        )
        assert [g.label for g in circ.gates] == [
            "U3",
            "CX",
            "U3",
            "CX",
            "U3",
        ] * 4

    def test_openqasm2_a_gate_called_gate(self):
        qasm_str = """
        OPENQASM 2.0;
        include "qelib1.inc";
        gate gate_PauliEvolution(param0) q0,q1 { rz(0.2) q0; rz(-0.1) q1; }
        qreg q[2];
        gate_PauliEvolution(0.1) q[0],q[1];
        """
        circ = qtn.Circuit.from_openqasm2_str(qasm_str)
        assert len(circ.gates) == 2

    @pytest.mark.parametrize(
        "Circ", [qtn.Circuit, qtn.CircuitMPS, qtn.CircuitDense]
    )
    def test_all_gate_methods(self, Circ):
        import random

        g_nq_np = [
            # single qubit
            ("x", 1, 0),
            ("y", 1, 0),
            ("z", 1, 0),
            ("s", 1, 0),
            ("sdg", 1, 0),
            ("t", 1, 0),
            ("tdg", 1, 0),
            ("h", 1, 0),
            ("sx", 1, 0),
            ("sxdg", 1, 0),
            ("iden", 1, 0),
            ("x_1_2", 1, 0),
            ("y_1_2", 1, 0),
            ("z_1_2", 1, 0),
            ("w_1_2", 1, 0),
            ("hz_1_2", 1, 0),
            # single qubit parametrizable
            ("rx", 1, 1),
            ("ry", 1, 1),
            ("rz", 1, 1),
            ("u3", 1, 3),
            ("u2", 1, 2),
            ("u1", 1, 1),
            ("phase", 1, 1),
            # two qubit
            ("cx", 2, 0),
            ("cy", 2, 0),
            ("cz", 2, 0),
            ("cnot", 2, 0),
            ("swap", 2, 0),
            ("iswap", 2, 0),
            # two qubit parametrizable
            ("rxx", 2, 1),
            ("ryy", 2, 1),
            ("rzz", 2, 1),
            ("crx", 2, 1),
            ("cry", 2, 1),
            ("crz", 2, 1),
            ("cu3", 2, 3),
            ("cu2", 2, 2),
            ("cu1", 2, 1),
            ("cphase", 2, 1),
            ("fsim", 2, 2),
            ("fsimg", 2, 5),
            ("givens", 2, 1),
            ("givens2", 2, 2),
            ("xx_plus_yy", 2, 2),
            ("xx_minus_yy", 2, 2),
            ("su4", 2, 15),
        ]
        random.shuffle(g_nq_np)

        psi0 = qtn.MPS_rand_state(2, 2)
        circ = Circ(2, psi0=psi0, tags="PSI0")

        for g, n_q, n_p in g_nq_np:
            args = [
                *np.random.uniform(0, 2 * np.pi, size=n_p),
                *np.random.choice([0, 1], replace=False, size=n_q),
            ]
            getattr(circ, g)(*args)

        assert circ.psi.H @ circ.psi == pytest.approx(1.0)
        assert abs((circ.psi.H & psi0) ^ all) < 0.99999999

    def test_su4(self):
        psi0 = qtn.MPS_rand_state(2, 2)
        circ_a = qtn.Circuit(psi0=psi0)
        params = qu.randn(15)

        circ_a.su4(*params, 0, 1)
        psi_a = circ_a.to_dense()

        circ_b = qtn.Circuit(psi0=psi0)
        (
            theta1,
            phi1,
            lamda1,
            theta2,
            phi2,
            lamda2,
            theta3,
            phi3,
            lamda3,
            theta4,
            phi4,
            lamda4,
            t1,
            t2,
            t3,
        ) = params
        circ_b.u3(theta1, phi1, lamda1, 0)
        circ_b.u3(theta2, phi2, lamda2, 1)
        circ_b.cnot(1, 0)
        circ_b.rz(t1, 0)
        circ_b.ry(t2, 1)
        circ_b.cnot(0, 1)
        circ_b.ry(t3, 1)
        circ_b.cnot(1, 0)
        circ_b.u3(theta3, phi3, lamda3, 0)
        circ_b.u3(theta4, phi4, lamda4, 1)
        psi_b = circ_b.to_dense()

        assert qu.fidelity(psi_a, psi_b) == pytest.approx(1.0)

    def test_three_qubit_gates(self):
        psi0 = qtn.MPS_rand_state(3, 2)
        circ = qtn.Circuit(psi0=psi0)
        circ.ccx(0, 1, 2)
        circ.cswap(2, 1, 0)
        circ.toffoli(0, 1, 2)
        circ.ccy(1, 0, 2)
        circ.ccz(1, 2, 0)
        circ.fredkin(2, 1, 0)
        psi = circ.psi.to_dense()
        assert qu.expec(psi, psi) == pytest.approx(1.0)

    def test_auto_split_gate(self):
        n = 3
        ops = [
            ("u3", 1.0, 2.0, 3.0, 0),
            ("u3", 2.0, 3.0, 1.0, 1),
            ("u3", 3.0, 1.0, 2.0, 2),
            ("cz", 0, 1),
            ("iswap", 1, 2),
            ("cx", 2, 0),
            ("iswap", 2, 1),
            ("h", 0),
            ("h", 1),
            ("h", 2),
        ]
        cnorm = qtn.Circuit(n, gate_opts=dict(contract="split-gate"))
        cnorm.apply_gates(ops)
        assert cnorm.psi.max_bond() == 4

        cswap = qtn.Circuit(n, gate_opts=dict(contract="swap-split-gate"))
        cswap.apply_gates(ops)
        assert cswap.psi.max_bond() == 4

        cauto = qtn.Circuit(n, gate_opts=dict(contract="auto-split-gate"))
        cauto.apply_gates(ops)
        assert cauto.psi.max_bond() == 2

        assert qu.fidelity(
            cnorm.psi.to_dense(), cswap.psi.to_dense()
        ) == pytest.approx(1.0)
        assert qu.fidelity(
            cswap.psi.to_dense(), cauto.psi.to_dense()
        ) == pytest.approx(1.0)

    @pytest.mark.parametrize("gate2", ["cx", "iswap"])
    def test_circuit_simplify_tensor_network(self, gate2):
        import itertools
        import random

        depth = n = 8

        circ = qtn.Circuit(n)

        def random_single_qubit_layer():
            return [
                (random.choice(["X_1_2", "Y_1_2", "W_1_2"]), i)
                for i in range(n)
            ]

        def even_two_qubit_layer():
            return [(gate2, i, i + 1) for i in range(0, n, 2)]

        def odd_two_qubit_layer():
            return [(gate2, i, i + 1) for i in range(1, n - 1, 2)]

        layering = itertools.cycle(
            [
                random_single_qubit_layer,
                even_two_qubit_layer,
                random_single_qubit_layer,
                odd_two_qubit_layer,
            ]
        )

        for i, layer_fn in zip(range(depth), layering):
            for g in layer_fn():
                circ.apply_gate(*g, gate_round=i)

        psif = qtn.MPS_computational_state("0" * n).squeeze_()
        tn = circ.psi & psif

        c = tn.contract(all)
        cw = tn.contraction_width()

        tn_s = tn.full_simplify()
        assert tn_s.num_tensors < tn.num_tensors
        assert tn_s.num_indices < tn.num_indices
        # need to specify output inds since we now have hyper edges
        c_s = tn_s.contract(all, output_inds=[])
        assert c_s == pytest.approx(c)
        cw_s = tn_s.contraction_width(output_inds=[])
        assert cw_s <= cw

    def test_amplitude(self):
        L = 5
        circ = random_a2a_circ(L, 3)
        psi = circ.to_dense()

        for i in range(2**L):
            b = f"{i:0>{L}b}"
            c = circ.amplitude(b)
            assert c == pytest.approx(psi[i, 0])

    def test_partial_trace(self):
        L = 5
        circ = random_a2a_circ(L, 3)
        psi = circ.to_dense()
        for i in range(L - 1):
            keep = (i, i + 1)
            assert_allclose(
                qu.partial_trace(psi, [2] * 5, keep=keep),
                circ.partial_trace(keep),
                atol=1e-12,
            )

    @pytest.mark.parametrize("group_size", (1, 2, 6))
    def test_sample(self, group_size):
        import collections

        from scipy.stats import power_divergence

        C = 2**10
        L = 5
        circ = random_a2a_circ(L, 3, seed=42)

        psi = circ.to_dense()
        p_exp = abs(psi.reshape(-1)) ** 2
        f_exp = p_exp * C

        counts = collections.Counter(
            circ.sample(C, group_size=group_size, seed=42)
        )
        f_obs = np.zeros(2**L)
        for b, c in counts.items():
            f_obs[int(b, 2)] = c

        assert power_divergence(f_obs, f_exp)[0] < 100

    @pytest.mark.parametrize("group_size", (1, 3))
    def test_sample_gate_by_gate(self, group_size):
        import collections

        from scipy.stats import power_divergence

        C = 2**10
        L = 5
        circ = random_a2a_circ(L, 3, seed=43)

        psi = circ.to_dense()
        p_exp = abs(psi.reshape(-1)) ** 2
        f_exp = p_exp * C

        counts = collections.Counter(
            circ.sample_gate_by_gate(C, group_size=group_size, seed=42)
        )
        f_obs = np.zeros(2**L)
        for b, c in counts.items():
            f_obs[int(b, 2)] = c

        assert power_divergence(f_obs, f_exp)[0] < 100

    def test_sample_chaotic(self):
        import collections

        from scipy.stats import power_divergence

        C = 2**12
        L = 5
        reps = 3
        depth = 2
        goodnesses = [0] * 5

        for i in range(reps):
            circ = random_a2a_circ(L, depth, seed=42 + i)

            psi = circ.to_dense()
            p_exp = abs(psi.reshape(-1)) ** 2
            f_exp = p_exp * C

            for num_marginal in [3, 4, 5]:
                counts = collections.Counter(
                    circ.sample_chaotic(C, num_marginal, seed=42 + i)
                )
                f_obs = np.zeros(2**L)
                for b, c in counts.items():
                    f_obs[int(b, 2)] = c

                goodness = power_divergence(f_obs, f_exp)[0]
                goodnesses[num_marginal - 1] += goodness

        # assert average sampling goodness gets better with larger marginal
        assert sum(goodnesses[i] < goodnesses[i - 1] for i in range(1, L)) == 2

    def test_local_expectation(self):
        import random

        L = 5
        depth = 3
        circ = random_a2a_circ(L, depth)
        psi = circ.to_dense()
        for _ in range(10):
            G = qu.rand_matrix(4)
            i = random.randint(0, L - 2)
            where = (i, i + 1)
            x1 = qu.expec(qu.ikron(G, [2] * L, where), psi)
            x2 = circ.local_expectation(G, where)
            assert x1 == pytest.approx(x2)

    def test_local_expectation_multigate(self):
        circ = qtn.Circuit(2)
        circ.h(0)
        circ.cnot(0, 1)
        circ.y(1)
        Gs = [qu.kronpow(qu.pauli(s), 2) for s in "xyz"]
        exps = circ.local_expectation(Gs, [0, 1])
        assert exps[0] == pytest.approx(-1)
        assert exps[1] == pytest.approx(-1)
        assert exps[2] == pytest.approx(-1)

    def test_local_expectation_len1(self):
        circ = qtn.Circuit(1)
        circ.apply_gate("H", 0, gate_round=0)
        circ.local_expectation([qu.pauli("X")], (0,))

    def test_uni_to_dense(self):
        import cmath

        circ = qft_circ(3)
        U = circ.uni.to_dense()
        w = cmath.exp(2j * math.pi / 2**3)
        ex = 2 ** (-3 / 2) * np.array(
            [
                [w**0, w**0, w**0, w**0, w**0, w**0, w**0, w**0],
                [w**0, w**1, w**2, w**3, w**4, w**5, w**6, w**7],
                [w**0, w**2, w**4, w**6, w**0, w**2, w**4, w**6],
                [w**0, w**3, w**6, w**1, w**4, w**7, w**2, w**5],
                [w**0, w**4, w**0, w**4, w**0, w**4, w**0, w**4],
                [w**0, w**5, w**2, w**7, w**4, w**1, w**6, w**3],
                [w**0, w**6, w**4, w**2, w**0, w**6, w**4, w**2],
                [w**0, w**7, w**6, w**5, w**4, w**3, w**2, w**1],
            ]
        )
        assert_allclose(U, ex)

    def test_swap_lighcones(self):
        circ = qtn.Circuit(3)
        circ.x(0)  # 0
        circ.x(1)  # 1
        circ.x(2)  # 2
        circ.swap(0, 1)  # 3
        circ.cx(1, 2)  # 4
        circ.cx(0, 1)  # 5
        assert circ.get_reverse_lightcone_tags((2,)) == (
            "PSI0",
            "GATE_0",
            "GATE_2",
            "GATE_4",
        )

    def test_swappy_local_expecs(self):
        circ = swappy_circ(4, 4)
        Gs = [qu.rand_matrix(4) for _ in range(3)]
        pairs = [(0, 1), (1, 2), (2, 3)]

        psi = circ.to_dense()
        dims = [2] * 4

        exs = [
            qu.expec(qu.ikron(G, dims, pair), psi)
            for G, pair in zip(Gs, pairs)
        ]
        aps = [circ.local_expectation(G, pair) for G, pair in zip(Gs, pairs)]

        assert_allclose(exs, aps)

    @pytest.mark.parametrize(
        "name, densefn, nparam, nqubit",
        [
            ("rx", qu.Rx, 1, 1),
            ("ry", qu.Ry, 1, 1),
            ("rz", qu.Rz, 1, 1),
            ("u3", qu.U_gate, 3, 1),
            ("fsim", qu.fsim, 2, 2),
            ("fsimg", qu.fsimg, 5, 2),
        ],
    )
    def test_parametrized_gates_rx(self, name, densefn, nparam, nqubit):
        k0 = qu.rand_ket(2**nqubit)
        params = qu.randn(nparam)
        kf = densefn(*params) @ k0
        k0mps = qtn.MatrixProductState.from_dense(k0, [2] * nqubit)
        circ = qtn.Circuit(psi0=k0mps, gate_opts={"contract": False})
        getattr(circ, name)(*params, *range(nqubit), parametrize=True)
        tn = circ.psi
        assert isinstance(tn["GATE_0"], qtn.PTensor)
        assert_allclose(circ.to_dense(), kf)

    def test_apply_raw_gate(self):
        k0 = qu.rand_ket(4)
        psi0 = qtn.MatrixProductState.from_dense(k0, [2] * 2)
        circ = qtn.Circuit(psi0=psi0)
        U = qu.rand_uni(4)
        circ.apply_gate_raw(U, [0, 1], tags="UCUSTOM")
        assert len(circ.gates) == 1
        assert "UCUSTOM" in circ.psi.tags
        assert qu.fidelity(circ.to_dense(), U @ k0) == pytest.approx(1)

    def test_apply_controlled_gate_basic_equiv(self):
        circ = qtn.Circuit(3)
        circ.apply_gate("x", qubits=(2,), controls=(0, 1))
        U = circ.get_uni().to_dense()
        assert_allclose(U, qu.toffoli())

        circ = qtn.Circuit(3)
        circ.apply_gate("swap", qubits=(1, 2), controls=(0,))
        U = circ.get_uni().to_dense()
        assert_allclose(U, qu.fredkin())

    def test_multi_controlled_circuit(self):
        import random

        N = 10
        circ = qtn.Circuit(N)
        regs = list(range(N))
        random.shuffle(regs)
        circ.apply_gate("H", regs[0])
        for i in range(N - 1):
            circ.apply_gate("CNOT", regs[i], regs[i + 1])
        circ.apply_gate("X", N - 1, controls=range(N - 1))
        circ.apply_gate("SWAP", qubits=(N - 2, N - 1), controls=range(N - 2))
        (b,) = circ.sample(1, group_size=3, seed=42)
        assert b[N - 2] == "0"

    @pytest.mark.parametrize("dtype", [None, "complex64", "complex128"])
    @pytest.mark.parametrize("backend", [None, "torch"])
    @pytest.mark.parametrize("dtype_final", [None, "complex64", "complex128"])
    @pytest.mark.parametrize("convert_eager", [True, False])
    def test_conversions(self, dtype, backend, dtype_final, convert_eager):
        if backend == "torch":
            pytest.importorskip("torch")

            def to_backend(x):
                import torch

                return torch.tensor(x)

        else:
            to_backend = None

        circ = qtn.Circuit(
            2, dtype=dtype, to_backend=to_backend, convert_eager=convert_eager
        )
        circ.h(0)
        circ.cx(0, 1)
        circ.y(1)

        if not convert_eager:
            # constructed with default dtype
            assert circ._psi.dtype_name == "complex128"
            assert circ._psi.backend == "numpy"
        else:
            # constructed with this type
            assert circ._psi.dtype == dtype or dtype is None
            if backend == "torch":
                assert circ._psi.backend == "torch"
            else:
                assert circ._psi.backend == "numpy"

        # converted to this type
        if dtype is None:
            expected_default_dtype = "complex128"
        else:
            expected_default_dtype = dtype

        if backend != "torch":
            test_tn_default = circ.amplitude_tn()
            test_tn_explicit = circ.amplitude_tn(dtype=dtype_final)
        else:
            # test a less simplified tensor network
            test_tn_default = circ.partial_trace_tn(
                (1,), simplify_sequence="R"
            )
            test_tn_explicit = circ.partial_trace_tn(
                (1,), simplify_sequence="R", dtype=dtype_final
            )

        assert test_tn_default.dtype_name == expected_default_dtype
        if dtype_final is not None:
            assert test_tn_explicit.dtype_name == dtype_final


class TestCircuitMPS:
    def test_from_qsim_mps_swapsplit(self):
        G = rand_reg_graph(reg=3, n=18, seed=42)
        qsim = graph_to_qsim(G)
        qc = qtn.CircuitMPS.from_qsim_str(qsim)
        assert len(qc.psi.tensors) == 18
        assert (qc.psi.H & qc.psi) ^ all == pytest.approx(1.0)

    def test_multi_controlled_mps_circuit(self):
        N = 10
        rng = np.random.default_rng(42)

        gates = []
        for i in range(N):
            gates.append(
                qtn.Gate(
                    "U3", params=rng.uniform(0, 2 * np.pi, size=3), qubits=[i]
                )
            )
        gates.append(
            qtn.Gate(
                "SU4",
                params=rng.uniform(0, 2 * np.pi, size=15),
                qubits=[6, 2],
                controls=[8, 3, 4, 0],
            )
        )
        for i in range(N):
            gates.append(
                qtn.Gate(
                    "U3", params=rng.uniform(0, 2 * np.pi, size=3), qubits=[i]
                )
            )
        gates.append(
            qtn.Gate.from_raw(
                qu.rand_uni(2**3), qubits=[0, 9, 5], controls=[1, 2, 7]
            )
        )

        circ = qtn.Circuit(N=10)
        circ.apply_gates(gates)
        psi_lazy = circ.psi
        circ = qtn.CircuitMPS(N=10)
        circ.apply_gates(gates)
        mps = circ.psi
        assert mps.norm() == pytest.approx(1.0)
        assert mps.distance_normalized(psi_lazy) < 1e-6

    def test_mps_sampling(self):
        N = 6
        circ = qtn.CircuitMPS(N)
        circ.h(3)
        circ.cx(3, 2)
        circ.cx(2, 1)
        circ.cx(1, 0)
        circ.cx(0, 5)
        circ.cx(5, 4)
        circ.x(4)
        for x in circ.sample(10, seed=42):
            assert x in {"000010", "111101"}

    def test_mps_sampling_seed(self):
        N = 1
        circ = qtn.CircuitMPS(N)
        circ.h(0)
        samples = list(circ.sample(10, seed=1234))
        assert len(set(samples)) == 2

    def test_permmps_sampling(self):
        N = 6
        circ = qtn.CircuitPermMPS(N)
        circ.h(3)
        circ.cx(3, 2)
        circ.cx(2, 1)
        circ.cx(1, 0)
        circ.cx(0, 5)
        circ.cx(5, 4)
        circ.x(4)
        assert circ.qubits != tuple(range(N))
        for x in circ.sample(10, seed=42):
            assert x in {"000010", "111101"}

    def test_permmps_sampling_seed(self):
        N = 1
        circ = qtn.CircuitPermMPS(N)
        circ.h(0)
        samples = list(circ.sample(10, seed=1234))
        assert len(set(samples)) == 2


class TestCircuitGen:
    @pytest.mark.parametrize(
        "ansatz,cyclic",
        [
            ("zigzag", False),
            ("brickwork", False),
            ("brickwork", True),
            ("rand", False),
            ("rand", True),
        ],
    )
    @pytest.mark.parametrize("n", [4, 5])
    def test_1D_ansatzes(self, ansatz, cyclic, n):
        depth = 3
        num_pairs = n if cyclic else n - 1

        fn = {
            "zigzag": qtn.circ_ansatz_1D_zigzag,
            "brickwork": qtn.circ_ansatz_1D_brickwork,
            "rand": qtn.circ_ansatz_1D_rand,
        }[ansatz]

        opts = dict(
            n=n,
            depth=3,
            gate_opts=dict(contract=False),
        )
        if cyclic:
            opts["cyclic"] = True
        if ansatz == "rand":
            opts["seed"] = 42

        circ = fn(**opts)
        tn = circ.uni

        # total number of entangling gates
        assert len(tn["CZ"]) == num_pairs * depth

        # number of entangling gates per pair
        for i in range(num_pairs):
            assert len(tn["CZ", f"I{i}", f"I{(i + 1) % n}"]) == depth

        assert all(isinstance(t, qtn.PTensor) for t in tn["U3"])

    def test_qaoa(self):
        G = rand_reg_graph(3, 10, seed=666)
        terms = {(i, j): 1.0 for i, j in G.edges}
        ZZ = qu.pauli("Z") & qu.pauli("Z")

        gammas = [-0.6]
        betas = [-0.4]

        circ1 = qtn.circ_qaoa(terms, 1, gammas, betas)

        energy1 = sum(circ1.local_expectation(ZZ, edge) for edge in terms)
        assert energy1 < -4

        gammas = [-0.4]
        betas = [0.3]

        circ2 = qtn.circ_qaoa(terms, 1, gammas, betas)

        energy2 = sum(circ2.local_expectation(ZZ, edge) for edge in terms)
        assert energy2 > 4
